@@ -22,6 +22,7 @@
 
 #include "partitions.h"
 #include "sysfs.h"
+#include "strutils.h"
 
 /**
  * SECTION: partitions
@@ -37,11 +38,13 @@
  *
  * @PTTYPE: partition table type (dos, gpt, etc.).
  *
+ * @PTUUID: partition table id (uuid for gpt, hex for dos).
+
  * @PART_ENTRY_SCHEME: partition table type
  *
  * @PART_ENTRY_NAME: partition name (gpt and mac only)
  *
- * @PART_ENTRY_UUID: partition UUID (gpt only)
+ * @PART_ENTRY_UUID: partition UUID (gpt, or pseudo IDs for MBR)
  *
  * @PART_ENTRY_TYPE: partition type, 0xNN (e.g 0x82) or type UUID (gpt only) or type string (mac)
  *
@@ -64,7 +67,7 @@
  *
  * pr = blkid_new_probe_from_filename(devname);
  * if (!pr)
- *	err("%s: faild to open device", devname);
+ *	err("%s: failed to open device", devname);
  *
  * blkid_probe_enable_partitions(pr, TRUE);
  * blkid_do_fullprobe(pr);
@@ -88,7 +91,7 @@
  *
  * pr = blkid_new_probe_from_filename(devname);
  * if (!pr)
- *	err("%s: faild to open device", devname);
+ *	err("%s: failed to open device", devname);
  *
  * ls = blkid_probe_get_partitions(pr);
  * nparts = blkid_partlist_numof_partitions(ls);
@@ -125,12 +128,14 @@ static const struct blkid_idinfo *idinfos[] =
 	&sun_pt_idinfo,
 	&dos_pt_idinfo,
 	&gpt_pt_idinfo,
+	&pmbr_pt_idinfo,	/* always after GPT */
 	&mac_pt_idinfo,
 	&ultrix_pt_idinfo,
 	&bsd_pt_idinfo,
 	&unixware_pt_idinfo,
 	&solaris_x86_pt_idinfo,
-	&minix_pt_idinfo
+	&minix_pt_idinfo,
+	&atari_pt_idinfo
 };
 
 /*
@@ -160,26 +165,27 @@ const struct blkid_chaindrv partitions_drv = {
 /* exported as opaque type "blkid_parttable" */
 struct blkid_struct_parttable {
 	const char	*type;		/* partition table type */
-	blkid_loff_t	offset;		/* begin of the partition table */
+	uint64_t	offset;		/* begin of the partition table (in bytes) */
 	int		nparts;		/* number of partitions */
 	blkid_partition	parent;		/* parent of nested partition table */
+	char		id[UUID_STR_LEN]; /* PT identifier (e.g. UUID for GPT) */
 
 	struct list_head t_tabs;	/* all tables */
 };
 
 /* exported as opaque type "blkid_partition" */
 struct blkid_struct_partition {
-	blkid_loff_t	start;		/* begin of the partition */
-	blkid_loff_t	size;		/* size of the partitions */
+	uint64_t	start;		/* begin of the partition (512-bytes sectors) */
+	uint64_t	size;		/* size of the partitions (512-bytes sectors) */
 
 	int		type;		/* partition type */
-	char		typestr[37];	/* partition type string (GPT and Mac) */
+	char		typestr[UUID_STR_LEN]; /* partition type string (GPT and Mac) */
 
 	unsigned long long flags;	/* partition flags / attributes */
 
 	int		partno;		/* partition number */
-	char		uuid[37];	/* UUID (when supported by PT), e.g GPT */
-	unsigned char	name[128];	/* Partition in UTF8 name (when supporte by PT), e.g. Mac */
+	char		uuid[UUID_STR_LEN]; /* UUID (when supported by PT), e.g GPT */
+	unsigned char	name[128];	/* Partition in UTF8 name (when supported by PT), e.g. Mac */
 
 	blkid_parttable	tab;		/* partition table */
 };
@@ -209,8 +215,6 @@ static int blkid_partitions_probe_partition(blkid_probe pr);
  */
 int blkid_probe_enable_partitions(blkid_probe pr, int enable)
 {
-	if (!pr)
-		return -1;
 	pr->chains[BLKID_CHAIN_PARTS].enabled = enable;
 	return 0;
 }
@@ -226,9 +230,6 @@ int blkid_probe_enable_partitions(blkid_probe pr, int enable)
  */
 int blkid_probe_set_partitions_flags(blkid_probe pr, int flags)
 {
-	if (!pr)
-		return -1;
-
 	pr->chains[BLKID_CHAIN_PARTS].flags = flags;
 	return 0;
 }
@@ -312,13 +313,16 @@ static void blkid_probe_set_partlist(blkid_probe pr, blkid_partlist ls)
 
 static void ref_parttable(blkid_parttable tab)
 {
-	tab->nparts++;
+	if (tab)
+		tab->nparts++;
 }
 
 static void unref_parttable(blkid_parttable tab)
 {
-	tab->nparts--;
+	if (!tab)
+		return;
 
+	tab->nparts--;
 	if (tab->nparts <= 0) {
 		list_del(&tab->t_tabs);
 		free(tab);
@@ -361,7 +365,7 @@ static void reset_partlist(blkid_partlist ls)
 	ls->next_partno = 1;
 	INIT_LIST_HEAD(&ls->l_tabs);
 
-	DBG(DEBUG_LOWPROBE, printf("partlist reset\n"));
+	DBG(LOWPROBE, ul_debug("partlist reset"));
 }
 
 static blkid_partlist partitions_init_data(struct blkid_chain *chn)
@@ -380,9 +384,7 @@ static blkid_partlist partitions_init_data(struct blkid_chain *chn)
 
 	reset_partlist(ls);
 
-	DBG(DEBUG_LOWPROBE,
-		printf("parts: initialized partitions list (%p, size=%d)\n",
-		ls, ls->nparts_max));
+	DBG(LOWPROBE, ul_debug("parts: initialized partitions list (size=%d)", ls->nparts_max));
 	return ls;
 }
 
@@ -402,7 +404,7 @@ static void partitions_free_data(blkid_probe pr __attribute__((__unused__)),
 }
 
 blkid_parttable blkid_partlist_new_parttable(blkid_partlist ls,
-				const char *type, blkid_loff_t offset)
+				const char *type, uint64_t offset)
 {
 	blkid_parttable tab;
 
@@ -416,9 +418,8 @@ blkid_parttable blkid_partlist_new_parttable(blkid_partlist ls,
 	INIT_LIST_HEAD(&tab->t_tabs);
 	list_add_tail(&tab->t_tabs, &ls->l_tabs);
 
-	DBG(DEBUG_LOWPROBE,
-		printf("parts: create a new partition table "
-		       "(%p, type=%s, offset=%"PRId64")\n", tab, type, offset));
+	DBG(LOWPROBE, ul_debug("parts: create a new partition table "
+		       "(type=%s, offset=%"PRId64")", type, offset));
 	return tab;
 }
 
@@ -428,12 +429,13 @@ static blkid_partition new_partition(blkid_partlist ls, blkid_parttable tab)
 
 	if (ls->nparts + 1 > ls->nparts_max) {
 		/* Linux kernel has DISK_MAX_PARTS=256, but it's too much for
-		 * generic Linux machine -- let start with 32 partititions.
+		 * generic Linux machine -- let start with 32 partitions.
 		 */
-		ls->parts = realloc(ls->parts, (ls->nparts_max + 32) *
+		void *tmp = realloc(ls->parts, (ls->nparts_max + 32) *
 					sizeof(struct blkid_struct_partition));
-		if (!ls->parts)
+		if (!tmp)
 			return NULL;
+		ls->parts = tmp;
 		ls->nparts_max += 32;
 	}
 
@@ -448,8 +450,7 @@ static blkid_partition new_partition(blkid_partlist ls, blkid_parttable tab)
 }
 
 blkid_partition blkid_partlist_add_partition(blkid_partlist ls,
-					blkid_parttable tab,
-					blkid_loff_t start, blkid_loff_t size)
+					blkid_parttable tab, uint64_t start, uint64_t size)
 {
 	blkid_partition par = new_partition(ls, tab);
 
@@ -459,14 +460,13 @@ blkid_partition blkid_partlist_add_partition(blkid_partlist ls,
 	par->start = start;
 	par->size = size;
 
-	DBG(DEBUG_LOWPROBE,
-		printf("parts: add partition (%p start=%"
-		PRId64 ", size=%" PRId64 ", table=%p)\n",
-		par, par->start, par->size, tab));
+	DBG(LOWPROBE, ul_debug("parts: add partition (start=%"
+		PRIu64 ", size=%" PRIu64 ")",
+		par->start, par->size));
 	return par;
 }
 
-/* allows to modify used partitions numbers (for example for logical partitions) */
+/* can be used to modify used partitions numbers (for example for logical partitions) */
 int blkid_partlist_set_partno(blkid_partlist ls, int partno)
 {
 	if (!ls)
@@ -480,8 +480,8 @@ int blkid_partlist_increment_partno(blkid_partlist ls)
 	return ls ? ls->next_partno++ : -1;
 }
 
-/* allows to set "parent" for the next nested partition */
-int blkid_partlist_set_parent(blkid_partlist ls, blkid_partition par)
+/* can be used to set "parent" for the next nested partition */
+static int blkid_partlist_set_parent(blkid_partlist ls, blkid_partition par)
 {
 	if (!ls)
 		return -1;
@@ -513,10 +513,10 @@ int blkid_partitions_get_flags(blkid_probe pr)
 
 /* check if @start and @size are within @par partition */
 int blkid_is_nested_dimension(blkid_partition par,
-			blkid_loff_t start, blkid_loff_t size)
+			uint64_t start, uint64_t size)
 {
-	blkid_loff_t pstart;
-	blkid_loff_t psize;
+	uint64_t pstart;
+	uint64_t psize;
 
 	if (!par)
 		return 0;
@@ -534,38 +534,42 @@ static int idinfo_probe(blkid_probe pr, const struct blkid_idinfo *id,
 			struct blkid_chain *chn)
 {
 	const struct blkid_idmag *mag = NULL;
-	blkid_loff_t off;
-	int rc = 1;		/* = nothing detected */
+	uint64_t off;
+	int rc = BLKID_PROBE_NONE;		/* default is nothing */
 
-	if (pr->size <= 0 || (id->minsz && id->minsz > pr->size))
+	if (pr->size <= 0 || (id->minsz && (unsigned)id->minsz > pr->size))
 		goto nothing;	/* the device is too small */
+	if (pr->flags & BLKID_FL_NOSCAN_DEV)
+		goto nothing;
 
-	if (blkid_probe_get_idmag(pr, id, &off, &mag))
+	rc = blkid_probe_get_idmag(pr, id, &off, &mag);
+	if (rc != BLKID_PROBE_OK)
 		goto nothing;
 
 	/* final check by probing function */
 	if (id->probefunc) {
-		DBG(DEBUG_LOWPROBE, printf(
-			"%s: ---> call probefunc()\n", id->name));
+		DBG(LOWPROBE, ul_debug(
+			"%s: ---> call probefunc()", id->name));
 		rc = id->probefunc(pr, mag);
-	        if (rc == -1) {
+		if (rc < 0) {
 			/* reset after error */
 			reset_partlist(blkid_probe_get_partlist(pr));
 			if (chn && !chn->binary)
-				blkid_probe_chain_reset_vals(pr, chn);
-			DBG(DEBUG_LOWPROBE, printf(
-				"%s probefunc failed\n", id->name));
+				blkid_probe_chain_reset_values(pr, chn);
+			DBG(LOWPROBE, ul_debug("%s probefunc failed, rc %d",
+						  id->name, rc));
 		}
-		if (rc == 0 && mag && chn && !chn->binary)
+		if (rc == BLKID_PROBE_OK && mag && chn && !chn->binary)
 			rc = blkid_probe_set_magic(pr, off, mag->len,
-					(unsigned char *) mag->magic);
+					(const unsigned char *) mag->magic);
 
-		DBG(DEBUG_LOWPROBE, printf(
-			"%s: <--- (rc = %d)\n", id->name, rc));
+		DBG(LOWPROBE, ul_debug("%s: <--- (rc = %d)", id->name, rc));
 	}
 
-nothing:
 	return rc;
+
+nothing:
+	return BLKID_PROBE_NONE;
 }
 
 /*
@@ -573,12 +577,16 @@ nothing:
  */
 static int partitions_probe(blkid_probe pr, struct blkid_chain *chn)
 {
-	int rc = 1;
+	int rc = BLKID_PROBE_NONE;
 	size_t i;
 
 	if (!pr || chn->idx < -1)
-		return -1;
-	blkid_probe_chain_reset_vals(pr, chn);
+		return -EINVAL;
+
+	blkid_probe_chain_reset_values(pr, chn);
+
+	if (pr->flags & BLKID_FL_NOSCAN_DEV)
+		return BLKID_PROBE_NONE;
 
 	if (chn->binary)
 		partitions_init_data(chn);
@@ -586,8 +594,7 @@ static int partitions_probe(blkid_probe pr, struct blkid_chain *chn)
 	if (!pr->wipe_size && (pr->prob_flags & BLKID_PROBE_FL_IGNORE_PT))
 		goto details_only;
 
-	DBG(DEBUG_LOWPROBE,
-		printf("--> starting probing loop [PARTS idx=%d]\n",
+	DBG(LOWPROBE, ul_debug("--> starting probing loop [PARTS idx=%d]",
 		chn->idx));
 
 	i = chn->idx < 0 ? 0 : chn->idx + 1U;
@@ -602,40 +609,53 @@ static int partitions_probe(blkid_probe pr, struct blkid_chain *chn)
 			continue;
 
 		/* apply checks from idinfo */
-		if (idinfo_probe(pr, idinfos[i], chn) != 0)
+		rc = idinfo_probe(pr, idinfos[i], chn);
+		if (rc < 0)
+			break;
+		if (rc != BLKID_PROBE_OK)
 			continue;
 
 		name = idinfos[i]->name;
 
-		/* all checks passed */
 		if (!chn->binary)
+			/*
+			 * Non-binary interface, set generic variables. Note
+			 * that the another variables could be set in prober
+			 * functions.
+			 */
 			blkid_probe_set_value(pr, "PTTYPE",
-						(unsigned char *) name,
+						(const unsigned char *) name,
 						strlen(name) + 1);
-		DBG(DEBUG_LOWPROBE,
-			printf("<-- leaving probing loop (type=%s) [PARTS idx=%d]\n",
+
+		DBG(LOWPROBE, ul_debug("<-- leaving probing loop (type=%s) [PARTS idx=%d]",
 			name, chn->idx));
-		rc = 0;
+		rc = BLKID_PROBE_OK;
 		break;
 	}
 
-	if (rc == 1) {
-		DBG(DEBUG_LOWPROBE,
-			printf("<-- leaving probing loop (failed) [PARTS idx=%d]\n",
-			chn->idx));
+	if (rc != BLKID_PROBE_OK) {
+		DBG(LOWPROBE, ul_debug("<-- leaving probing loop (failed=%d) [PARTS idx=%d]",
+			rc, chn->idx));
 	}
 
 details_only:
 	/*
 	 * Gather PART_ENTRY_* values if the current device is a partition.
 	 */
-	if (!chn->binary &&
+	if ((rc == BLKID_PROBE_OK || rc == BLKID_PROBE_NONE) && !chn->binary &&
 	    (blkid_partitions_get_flags(pr) & BLKID_PARTS_ENTRY_DETAILS)) {
 
-		if (!blkid_partitions_probe_partition(pr))
-			rc = 0;
+		int xrc = blkid_partitions_probe_partition(pr);
+
+		/* partition entry probing is optional, and "not-found" from
+		 * this sub-probing must not to overwrite previous success. */
+		if (xrc < 0)
+			rc = xrc;			/* always propagate errors */
+		else if (rc == BLKID_PROBE_NONE)
+			rc = xrc;
 	}
 
+	DBG(LOWPROBE, ul_debug("partitions probe done [rc=%d]",	rc));
 	return rc;
 }
 
@@ -644,32 +664,34 @@ int blkid_partitions_do_subprobe(blkid_probe pr, blkid_partition parent,
 		const struct blkid_idinfo *id)
 {
 	blkid_probe prc;
-	int rc = 1;
+	int rc;
 	blkid_partlist ls;
-	blkid_loff_t sz, off;
+	uint64_t sz, off;
 
-	DBG(DEBUG_LOWPROBE, printf(
-		"parts: ----> %s subprobe requested (parent=%p)\n",
-		id->name, parent));
+	DBG(LOWPROBE, ul_debug(
+		"parts: ----> %s subprobe requested)",
+		id->name));
 
 	if (!pr || !parent || !parent->size)
-		return -1;
+		return -EINVAL;
+	if (pr->flags & BLKID_FL_NOSCAN_DEV)
+		return BLKID_PROBE_NONE;
 
 	/* range defined by parent */
-	sz = ((blkid_loff_t) parent->size) << 9;
-	off = ((blkid_loff_t) parent->start) << 9;
+	sz = parent->size << 9;
+	off = parent->start << 9;
 
 	if (off < pr->off || pr->off + pr->size < off + sz) {
-		DBG(DEBUG_LOWPROBE, printf(
-			"ERROR: parts: <---- '%s' subprobe: overflow detected.\n",
+		DBG(LOWPROBE, ul_debug(
+			"ERROR: parts: <---- '%s' subprobe: overflow detected.",
 			id->name));
-		return -1;
+		return -ENOSPC;
 	}
 
 	/* create private prober */
 	prc = blkid_clone_probe(pr);
 	if (!prc)
-		return -1;
+		return -ENOMEM;
 
 	blkid_probe_set_dimension(prc, off, sz);
 
@@ -693,21 +715,25 @@ int blkid_partitions_do_subprobe(blkid_probe pr, blkid_partition parent,
 
 	blkid_free_probe(prc);	/* free cloned prober */
 
-	DBG(DEBUG_LOWPROBE, printf(
-		"parts: <---- %s subprobe done (parent=%p, rc=%d)\n",
-		id->name, parent, rc));
+	DBG(LOWPROBE, ul_debug(
+		"parts: <---- %s subprobe done (rc=%d)",
+		id->name, rc));
 
 	return rc;
 }
 
 static int blkid_partitions_probe_partition(blkid_probe pr)
 {
-	int rc = 1;
 	blkid_probe disk_pr = NULL;
 	blkid_partlist ls;
 	blkid_partition par;
 	dev_t devno;
 
+	DBG(LOWPROBE, ul_debug("parts: start probing for partition entry"));
+
+	if (pr->flags & BLKID_FL_NOSCAN_DEV)
+		goto nothing;
+
 	devno = blkid_probe_get_devno(pr);
 	if (!devno)
 		goto nothing;
@@ -722,7 +748,9 @@ static int blkid_partitions_probe_partition(blkid_probe pr)
 		goto nothing;
 
 	par = blkid_partlist_devno_to_partition(ls, devno);
-	if (par) {
+	if (!par)
+		goto nothing;
+	else {
 		const char *v;
 		blkid_parttable tab = blkid_partition_get_table(par);
 		dev_t disk = blkid_probe_get_devno(disk_pr);
@@ -731,24 +759,24 @@ static int blkid_partitions_probe_partition(blkid_probe pr)
 			v = blkid_parttable_get_type(tab);
 			if (v)
 				blkid_probe_set_value(pr, "PART_ENTRY_SCHEME",
-					(unsigned char *) v, strlen(v) + 1);
+					(const unsigned char *) v, strlen(v) + 1);
 		}
 
 		v = blkid_partition_get_name(par);
 		if (v)
 			blkid_probe_set_value(pr, "PART_ENTRY_NAME",
-				(unsigned char *) v, strlen(v) + 1);
+				(const unsigned char *) v, strlen(v) + 1);
 
 		v = blkid_partition_get_uuid(par);
 		if (v)
 			blkid_probe_set_value(pr, "PART_ENTRY_UUID",
-				(unsigned char *) v, strlen(v) + 1);
+				(const unsigned char *) v, strlen(v) + 1);
 
 		/* type */
 		v = blkid_partition_get_type_string(par);
 		if (v)
 			blkid_probe_set_value(pr, "PART_ENTRY_TYPE",
-				(unsigned char *) v, strlen(v) + 1);
+				(const unsigned char *) v, strlen(v) + 1);
 		else
 			blkid_probe_sprintf_value(pr, "PART_ENTRY_TYPE",
 				"0x%x", blkid_partition_get_type(par));
@@ -761,16 +789,22 @@ static int blkid_partitions_probe_partition(blkid_probe pr)
 				"%d", blkid_partition_get_partno(par));
 
 		blkid_probe_sprintf_value(pr, "PART_ENTRY_OFFSET", "%jd",
-				blkid_partition_get_start(par));
+				(intmax_t)blkid_partition_get_start(par));
 		blkid_probe_sprintf_value(pr, "PART_ENTRY_SIZE", "%jd",
-				blkid_partition_get_size(par));
+				(intmax_t)blkid_partition_get_size(par));
 
 		blkid_probe_sprintf_value(pr, "PART_ENTRY_DISK", "%u:%u",
 				major(disk), minor(disk));
 	}
-	rc = 0;
+
+	DBG(LOWPROBE, ul_debug("parts: end probing for partition entry [success]"));
+	return BLKID_PROBE_OK;
+
 nothing:
-	return rc;
+	DBG(LOWPROBE, ul_debug("parts: end probing for partition entry [nothing]"));
+	return BLKID_PROBE_NONE;
+
+
 }
 
 /*
@@ -778,17 +812,20 @@ nothing:
  * @size is covered by any partition.
  */
 int blkid_probe_is_covered_by_pt(blkid_probe pr,
-				 blkid_loff_t offset, blkid_loff_t size)
+				 uint64_t offset, uint64_t size)
 {
-	blkid_probe prc;
+	blkid_probe prc = NULL;
 	blkid_partlist ls = NULL;
-	blkid_loff_t start, end;
+	uint64_t start, end;
 	int nparts, i, rc = 0;
 
-	DBG(DEBUG_LOWPROBE, printf(
-		"=> checking if off=%jd size=%jd covered by PT\n",
+	DBG(LOWPROBE, ul_debug(
+		"=> checking if off=%"PRIu64" size=%"PRIu64" covered by PT",
 		offset, size));
 
+	if (pr->flags & BLKID_FL_NOSCAN_DEV)
+		goto done;
+
 	prc = blkid_clone_probe(pr);
 	if (!prc)
 		goto done;
@@ -809,8 +846,8 @@ int blkid_probe_is_covered_by_pt(blkid_probe pr,
 		blkid_partition par = &ls->parts[i];
 
 		if (par->start + par->size > (pr->size >> 9)) {
-			DBG(DEBUG_LOWPROBE, printf("partition #%d overflows "
-				"device (off=%" PRId64 " size=%" PRId64 ")\n",
+			DBG(LOWPROBE, ul_debug("partition #%d overflows "
+				"device (off=%" PRId64 " size=%" PRId64 ")",
 				par->partno, par->start, par->size));
 			goto done;
 		}
@@ -828,13 +865,13 @@ int blkid_probe_is_covered_by_pt(blkid_probe pr,
 done:
 	blkid_free_probe(prc);
 
-	DBG(DEBUG_LOWPROBE, printf("<= %s covered by PT\n", rc ? "IS" : "NOT"));
+	DBG(LOWPROBE, ul_debug("<= %s covered by PT", rc ? "IS" : "NOT"));
 	return rc;
 }
 
 /**
  * blkid_known_pttype:
- * @pttype: partiton name
+ * @pttype: partition name
  *
  * Returns: 1 for known or 0 for unknown partition type.
  */
@@ -853,6 +890,24 @@ int blkid_known_pttype(const char *pttype)
 	return 0;
 }
 
+/**
+ * blkid_partitions_get_name:
+ * @idx: number >= 0
+ * @name: returns name of a supported partition
+ *
+ * Since: 2.30
+ *
+ * Returns: -1 if @idx is out of range, or 0 on success.
+ */
+int blkid_partitions_get_name(const size_t idx, const char **name)
+{
+	if (idx < ARRAY_SIZE(idinfos)) {
+		*name = idinfos[idx]->name;
+		return 0;
+	}
+	return -1;
+}
+
 /**
  * blkid_partlist_numof_partitions:
  * @ls: partitions list
@@ -861,7 +916,7 @@ int blkid_known_pttype(const char *pttype)
  */
 int blkid_partlist_numof_partitions(blkid_partlist ls)
 {
-	return ls ? ls->nparts : -1;
+	return ls->nparts;
 }
 
 /**
@@ -873,7 +928,7 @@ int blkid_partlist_numof_partitions(blkid_partlist ls)
  */
 blkid_parttable blkid_partlist_get_table(blkid_partlist ls)
 {
-	if (!ls || list_empty(&ls->l_tabs))
+	if (list_empty(&ls->l_tabs))
 		return NULL;
 
 	return list_entry(ls->l_tabs.next,
@@ -896,12 +951,52 @@ blkid_parttable blkid_partlist_get_table(blkid_partlist ls)
  */
 blkid_partition blkid_partlist_get_partition(blkid_partlist ls, int n)
 {
-	if (!ls || n < 0 || n >= ls->nparts)
+	if (n < 0 || n >= ls->nparts)
 		return NULL;
 
 	return &ls->parts[n];
 }
 
+blkid_partition blkid_partlist_get_partition_by_start(blkid_partlist ls, uint64_t start)
+{
+	int i, nparts;
+	blkid_partition par;
+
+	nparts = blkid_partlist_numof_partitions(ls);
+	for (i = 0; i < nparts; i++) {
+		par = blkid_partlist_get_partition(ls, i);
+		if ((uint64_t) blkid_partition_get_start(par) == start)
+			return par;
+	}
+	return NULL;
+}
+
+/**
+ * blkid_partlist_get_partition_by_partno
+ * @ls: partitions list
+ * @n: the partition number (e.g. 'N' from sda'N')
+ *
+ * This does not assume any order of the input blkid_partlist.  And correctly
+ * handles "out of order" partition tables.  partition N is located after
+ * partition N+1 on the disk.
+ *
+ * Returns: partition object or NULL in case or error.
+ */
+blkid_partition blkid_partlist_get_partition_by_partno(blkid_partlist ls, int n)
+{
+	int i, nparts;
+	blkid_partition par;
+
+	nparts = blkid_partlist_numof_partitions(ls);
+	for (i = 0; i < nparts; i++) {
+		par = blkid_partlist_get_partition(ls, i);
+		if (n == blkid_partition_get_partno(par))
+			return par;
+	}
+	return NULL;
+}
+
+
 /**
  * blkid_partlist_devno_to_partition:
  * @ls: partitions list
@@ -917,27 +1012,30 @@ blkid_partition blkid_partlist_get_partition(blkid_partlist ls, int n)
  */
 blkid_partition blkid_partlist_devno_to_partition(blkid_partlist ls, dev_t devno)
 {
-	struct sysfs_cxt sysfs;
-	uint64_t start, size;
+	struct path_cxt *pc;
+	uint64_t start = 0, size;
 	int i, rc, partno = 0;
 
-	DBG(DEBUG_LOWPROBE,
-		printf("triyng to convert devno 0x%llx to partition\n",
+	DBG(LOWPROBE, ul_debug("trying to convert devno 0x%llx to partition",
 			(long long) devno));
 
-	if (sysfs_init(&sysfs, devno, NULL)) {
-		DBG(DEBUG_LOWPROBE, printf("failed t init sysfs context\n"));
+
+	pc = ul_new_sysfs_path(devno, NULL, NULL);
+	if (!pc) {
+		DBG(LOWPROBE, ul_debug("failed t init sysfs context"));
 		return NULL;
 	}
-	rc = sysfs_read_u64(&sysfs, "size", &size);
+	rc = ul_path_read_u64(pc, &size, "size");
 	if (!rc) {
-		rc = sysfs_read_u64(&sysfs, "start", &start);
+		rc = ul_path_read_u64(pc, &start, "start");
 		if (rc) {
 			/* try to get partition number from DM uuid.
 			 */
-			char *uuid = sysfs_strdup(&sysfs, "dm/uuid");
-			char *tmp = uuid;
-			char *prefix = uuid ? strsep(&tmp, "-") : NULL;
+			char *uuid = NULL, *tmp, *prefix;
+
+			ul_path_read_string(pc, &uuid, "dm/uuid");
+			tmp = uuid;
+			prefix = uuid ? strsep(&tmp, "-") : NULL;
 
 			if (prefix && strncasecmp(prefix, "part", 4) == 0) {
 				char *end = NULL;
@@ -952,18 +1050,18 @@ blkid_partition blkid_partlist_devno_to_partition(blkid_partlist ls, dev_t devno
 		}
 	}
 
-	sysfs_deinit(&sysfs);
+	ul_unref_path(pc);
 
 	if (rc)
 		return NULL;
 
 	if (partno) {
-		DBG(DEBUG_LOWPROBE, printf("mapped by DM, using partno %d\n", partno));
+		DBG(LOWPROBE, ul_debug("mapped by DM, using partno %d", partno));
 
 		/*
 		 * Partition mapped by kpartx does not provide "start" offset
 		 * in /sys, but if we know partno and size of the partition
-		 * that we can probably make the releation bettween the device
+		 * that we can probably make the relation between the device
 		 * and an entry in partition table.
 		 */
 		 for (i = 0; i < ls->nparts; i++) {
@@ -972,38 +1070,108 @@ blkid_partition blkid_partlist_devno_to_partition(blkid_partlist ls, dev_t devno
 			 if (partno != blkid_partition_get_partno(par))
 				 continue;
 
-			 if ((blkid_loff_t) size == blkid_partition_get_size(par) ||
-			     (blkid_partition_is_extended(par) && size <= 1024))
+			 if (size == (uint64_t)blkid_partition_get_size(par) ||
+			     (blkid_partition_is_extended(par) && size <= 1024ULL))
 				 return par;
 
 		 }
 		 return NULL;
 	}
 
-	DBG(DEBUG_LOWPROBE, printf("searching by offset/size\n"));
+	DBG(LOWPROBE, ul_debug("searching by offset/size"));
 
 	for (i = 0; i < ls->nparts; i++) {
 		blkid_partition par = &ls->parts[i];
 
-		if (blkid_partition_get_start(par) == (blkid_loff_t) start &&
-		    blkid_partition_get_size(par) == (blkid_loff_t) size)
+		if ((uint64_t)blkid_partition_get_start(par) == start &&
+		    (uint64_t)blkid_partition_get_size(par) == size)
 			return par;
 
 		/* exception for extended dos partitions */
-		if (blkid_partition_get_start(par) == (blkid_loff_t) start &&
-		    blkid_partition_is_extended(par) && size <= 1024)
+		if ((uint64_t)blkid_partition_get_start(par) == start &&
+		    blkid_partition_is_extended(par) && size <= 1024ULL)
 			return par;
 
 	}
 
-	DBG(DEBUG_LOWPROBE, printf("not found partition for device\n"));
+	DBG(LOWPROBE, ul_debug("not found partition for device"));
 	return NULL;
 }
 
+
+int blkid_parttable_set_uuid(blkid_parttable tab, const unsigned char *id)
+{
+	if (!tab)
+		return -1;
+
+	blkid_unparse_uuid(id, tab->id, sizeof(tab->id));
+	return 0;
+}
+
+int blkid_parttable_set_id(blkid_parttable tab, const unsigned char *id)
+{
+	if (!tab)
+		return -1;
+
+	xstrncpy(tab->id, (const char *) id, sizeof(tab->id));
+	return 0;
+}
+
+/* set PTUUID variable for non-binary API */
+int blkid_partitions_set_ptuuid(blkid_probe pr, unsigned char *uuid)
+{
+	struct blkid_chain *chn = blkid_probe_get_chain(pr);
+	struct blkid_prval *v;
+
+	if (chn->binary || blkid_uuid_is_empty(uuid, 16))
+		return 0;
+
+	v = blkid_probe_assign_value(pr, "PTUUID");
+	if (!v)
+		return -ENOMEM;
+
+	v->len = UUID_STR_LEN;
+	v->data = calloc(1, v->len);
+	if (v->data) {
+		blkid_unparse_uuid(uuid, (char *) v->data, v->len);
+		return 0;
+	}
+
+	blkid_probe_free_value(v);
+	return -ENOMEM;
+}
+
+/* set PTUUID variable for non-binary API for tables where
+ * the ID is just a string */
+int blkid_partitions_strcpy_ptuuid(blkid_probe pr, char *str)
+{
+	struct blkid_chain *chn = blkid_probe_get_chain(pr);
+
+	if (chn->binary || !str || !*str)
+		return 0;
+
+	if (!blkid_probe_set_value(pr, "PTUUID", (unsigned char *) str, strlen(str) + 1))
+		return -ENOMEM;
+
+	return 0;
+}
+
+/**
+ * blkid_parttable_get_id:
+ * @tab: partition table
+ *
+ * The ID is GPT disk UUID or DOS disk ID (in hex format).
+ *
+ * Returns: partition table ID (for example GPT disk UUID) or NULL
+ */
+const char *blkid_parttable_get_id(blkid_parttable tab)
+{
+	return *tab->id ? tab->id : NULL;
+}
+
+
 int blkid_partition_set_type(blkid_partition par, int type)
 {
-	if (!par)
-		return -1;
 	par->type = type;
 	return 0;
 }
@@ -1016,18 +1184,18 @@ int blkid_partition_set_type(blkid_partition par, int type)
  */
 const char *blkid_parttable_get_type(blkid_parttable tab)
 {
-	return tab ? tab->type : NULL;
+	return tab->type;
 }
 
 /**
  * blkid_parttable_get_parent:
  * @tab: partition table
  *
- * Returns: parent for nexted partitition tables or NULL.
+ * Returns: parent for nested partition tables or NULL.
  */
 blkid_partition blkid_parttable_get_parent(blkid_parttable tab)
 {
-	return tab ? tab->parent : NULL;
+	return tab->parent;
 }
 
 /**
@@ -1036,7 +1204,7 @@ blkid_partition blkid_parttable_get_parent(blkid_parttable tab)
  *
  * Note the position is relative to begin of the device as defined by
  * blkid_probe_set_device() for primary partition table, and relative
- * to parental partition for nested patition tables.
+ * to parental partition for nested partition tables.
  *
  * <informalexample>
  *   <programlisting>
@@ -1056,7 +1224,7 @@ blkid_partition blkid_parttable_get_parent(blkid_parttable tab)
  */
 blkid_loff_t blkid_parttable_get_offset(blkid_parttable tab)
 {
-	return tab ? tab->offset : -1;
+	return (blkid_loff_t)tab->offset;
 }
 
 /**
@@ -1092,7 +1260,7 @@ blkid_loff_t blkid_parttable_get_offset(blkid_parttable tab)
  */
 blkid_parttable blkid_partition_get_table(blkid_partition par)
 {
-	return par ? par->tab : NULL;
+	return par->tab;
 }
 
 static int partition_get_logical_type(blkid_partition par)
@@ -1113,9 +1281,9 @@ static int partition_get_logical_type(blkid_partition par)
 		if (par->partno > 4)
 			return 'L';	/* logical */
 
-	        if(par->type == BLKID_DOS_EXTENDED_PARTITION ||
-                   par->type == BLKID_W95_EXTENDED_PARTITION ||
-		   par->type == BLKID_LINUX_EXTENDED_PARTITION)
+	        if(par->type == MBR_DOS_EXTENDED_PARTITION ||
+                   par->type == MBR_W95_EXTENDED_PARTITION ||
+		   par->type == MBR_LINUX_EXTENDED_PARTITION)
 			return 'E';
 	}
 	return 'P';
@@ -1189,7 +1357,7 @@ int blkid_partition_set_utf8name(blkid_partition par, const unsigned char *name,
 	if (!par)
 		return -1;
 
-	blkid_encode_to_utf8(enc, par->name, sizeof(par->name), name, len);
+	ul_encode_to_utf8(enc, par->name, sizeof(par->name), name, len);
 	blkid_rtrim_whitespace(par->name);
 	return 0;
 }
@@ -1203,6 +1371,16 @@ int blkid_partition_set_uuid(blkid_partition par, const unsigned char *uuid)
 	return 0;
 }
 
+int blkid_partition_gen_uuid(blkid_partition par)
+{
+	if (!par || !par->tab || !*par->tab->id)
+		return -1;
+
+	snprintf(par->uuid, sizeof(par->uuid), "%.33s-%02x",
+			par->tab->id, par->partno);
+	return 0;
+}
+
 /**
  * blkid_partition_get_name:
  * @par: partition
@@ -1211,7 +1389,7 @@ int blkid_partition_set_uuid(blkid_partition par, const unsigned char *uuid)
  */
 const char *blkid_partition_get_name(blkid_partition par)
 {
-	return par && *par->name ? (char *) par->name : NULL;
+	return *par->name ? (char *) par->name : NULL;
 }
 
 /**
@@ -1222,19 +1400,19 @@ const char *blkid_partition_get_name(blkid_partition par)
  */
 const char *blkid_partition_get_uuid(blkid_partition par)
 {
-	return par && *par->uuid ? par->uuid : NULL;
+	return *par->uuid ? par->uuid : NULL;
 }
 
 /**
  * blkid_partition_get_partno:
  * @par: partition
  *
- * Returns: proposed partitin number (e.g. 'N' from sda'N') or -1 in case of
- * error. Note that the number is generate by library independenly on your OS.
+ * Returns: proposed partition number (e.g. 'N' from sda'N') or -1 in case of
+ * error. Note that the number is generate by library independently on your OS.
  */
 int blkid_partition_get_partno(blkid_partition par)
 {
-	return par ? par->partno : -1;
+	return par->partno;
 }
 
 /**
@@ -1243,21 +1421,21 @@ int blkid_partition_get_partno(blkid_partition par)
  *
  * Be careful if you _not_ probe whole disk:
  *
- * 1) the offset is usully relative to begin of the disk -- but if you probe a
+ * 1) the offset is usually relative to begin of the disk -- but if you probe a
  *    fragment of the disk only -- then the offset could be still relative to
  *    the begin of the disk rather that relative to the fragment.
  *
- * 2) the offset for nested partitions could be releative to parent (e.g. Solaris)
+ * 2) the offset for nested partitions could be relative to parent (e.g. Solaris)
  *    _or_ relative to the begin of the whole disk (e.g. bsd).
  *
- * You don't have to care about such details if you proble whole disk. In such
+ * You don't have to care about such details if you probe whole disk. In such
  * a case libblkid always returns the offset relative to the begin of the disk.
  *
  * Returns: start of the partition (in 512-sectors).
  */
 blkid_loff_t blkid_partition_get_start(blkid_partition par)
 {
-	return par ? par->start : -1;
+	return (blkid_loff_t)par->start;
 }
 
 /**
@@ -1268,7 +1446,7 @@ blkid_loff_t blkid_partition_get_start(blkid_partition par)
  *          library always returns full size of the partition. If you want add
  *          the partition to the Linux system (BLKPG_ADD_PARTITION ioctl) you
  *          need to reduce the size of the partition to 1 or 2 blocks. The
- *          rest of the partition has to be unaccessible for mkfs or mkswap
+ *          rest of the partition has to be inaccessible for mkfs or mkswap
  *          programs, we need a small space for boot loaders only.
  *
  *          For some unknown reason this (safe) practice is not to used for
@@ -1278,7 +1456,7 @@ blkid_loff_t blkid_partition_get_start(blkid_partition par)
  */
 blkid_loff_t blkid_partition_get_size(blkid_partition par)
 {
-	return par ? par->size : -1;
+	return (blkid_loff_t)par->size;
 }
 
 /**
@@ -1289,7 +1467,7 @@ blkid_loff_t blkid_partition_get_size(blkid_partition par)
  */
 int blkid_partition_get_type(blkid_partition par)
 {
-	return par ? par->type : 0;
+	return par->type;
 }
 
 /* Sets partition 'type' for PT where the type is defined by string rather
@@ -1298,9 +1476,6 @@ int blkid_partition_get_type(blkid_partition par)
 int blkid_partition_set_type_string(blkid_partition par,
 		const unsigned char *type, size_t len)
 {
-	if (!par)
-		return -1;
-
 	set_string((unsigned char *) par->typestr,
 			sizeof(par->typestr), type, len);
 	return 0;
@@ -1311,9 +1486,6 @@ int blkid_partition_set_type_string(blkid_partition par,
  */
 int blkid_partition_set_type_uuid(blkid_partition par, const unsigned char *uuid)
 {
-	if (!par)
-		return -1;
-
 	blkid_unparse_uuid(uuid, par->typestr, sizeof(par->typestr));
 	return 0;
 }
@@ -1330,14 +1502,12 @@ int blkid_partition_set_type_uuid(blkid_partition par, const unsigned char *uuid
  */
 const char *blkid_partition_get_type_string(blkid_partition par)
 {
-	return par && *par->typestr ? par->typestr : NULL;
+	return *par->typestr ? par->typestr : NULL;
 }
 
 
 int blkid_partition_set_flags(blkid_partition par, unsigned long long flags)
 {
-	if (!par)
-		return -1;
 	par->flags = flags;
 	return 0;
 }
@@ -1350,6 +1520,6 @@ int blkid_partition_set_flags(blkid_partition par, unsigned long long flags)
  */
 unsigned long long blkid_partition_get_flags(blkid_partition par)
 {
-	return par ? par->flags : 0;
+	return par->flags;
 }
 
