@@ -105,10 +105,6 @@
 #include <stdarg.h>
 #include <limits.h>
 
-#ifdef HAVE_LIBUUID
-# include <uuid.h>
-#endif
-
 #include "blkidP.h"
 #include "all-io.h"
 #include "sysfs.h"
@@ -124,9 +120,7 @@ static const struct blkid_chaindrv *chains_drvs[] = {
 	[BLKID_CHAIN_PARTS] = &partitions_drv
 };
 
-static struct blkid_prval *blkid_probe_new_value(void);
 static void blkid_probe_reset_values(blkid_probe pr);
-static void blkid_probe_reset_buffer(blkid_probe pr);
 
 /**
  * blkid_new_probe:
@@ -143,7 +137,7 @@ blkid_probe blkid_new_probe(void)
 	if (!pr)
 		return NULL;
 
-	DBG(LOWPROBE, ul_debug("allocate a new probe %p", pr));
+	DBG(LOWPROBE, ul_debug("allocate a new probe"));
 
 	/* initialize chains */
 	for (i = 0; i < BLKID_NCHAINS; i++) {
@@ -206,10 +200,10 @@ blkid_probe blkid_clone_probe(blkid_probe parent)
  */
 blkid_probe blkid_new_probe_from_filename(const char *filename)
 {
-	int fd = -1;
+	int fd;
 	blkid_probe pr = NULL;
 
-	fd = open(filename, O_RDONLY|O_CLOEXEC);
+	fd = open(filename, O_RDONLY|O_CLOEXEC|O_NONBLOCK);
 	if (fd < 0)
 		return NULL;
 
@@ -223,8 +217,7 @@ blkid_probe blkid_new_probe_from_filename(const char *filename)
 	pr->flags |= BLKID_FL_PRIVATE_FD;
 	return pr;
 err:
-	if (fd >= 0)
-		close(fd);
+	close(fd);
 	blkid_free_probe(pr);
 	return NULL;
 }
@@ -253,11 +246,11 @@ void blkid_free_probe(blkid_probe pr)
 
 	if ((pr->flags & BLKID_FL_PRIVATE_FD) && pr->fd >= 0)
 		close(pr->fd);
-	blkid_probe_reset_buffer(pr);
+	blkid_probe_reset_buffers(pr);
 	blkid_probe_reset_values(pr);
 	blkid_free_probe(pr->disk_probe);
 
-	DBG(LOWPROBE, ul_debug("free probe %p", pr));
+	DBG(LOWPROBE, ul_debug("free probe"));
 	free(pr);
 }
 
@@ -300,27 +293,6 @@ static void blkid_probe_chain_reset_position(struct blkid_chain *chn)
 	chn->idx = -1;
 }
 
-/*
-static struct blkid_prval *blkid_probe_copy_value(struct blkid_prval *src)
-{
-	struct blkid_prval *dest = blkid_probe_new_value();
-
-	if (!dest)
-		return NULL;
-
-	memcpy(dest, src, sizeof(struct blkid_prval));
-
-	dest->data = malloc(src->len);
-	if (!dest->data)
-		return NULL;
-
-	memcpy(dest->data, src->data, src->len);
-
-	INIT_LIST_HEAD(&dest->prvals);
-	return dest;
-}
-*/
-
 /*
  * Move chain values from probing result to @vals
  */
@@ -338,9 +310,7 @@ int blkid_probe_chain_save_values(blkid_probe pr, struct blkid_chain *chn,
 		if (v->chain != chn)
 			continue;
 
-		list_del(&v->prvals);
-		INIT_LIST_HEAD(&v->prvals);
-
+		list_del_init(&v->prvals);
 		list_add_tail(&v->prvals, vals);
 	}
 	return 0;
@@ -565,7 +535,7 @@ static struct blkid_bufinfo *read_buffer(blkid_probe pr, uint64_t real_off, uint
 	ssize_t ret;
 	struct blkid_bufinfo *bf = NULL;
 
-	if (blkid_llseek(pr->fd, real_off, SEEK_SET) < 0) {
+	if (lseek(pr->fd, real_off, SEEK_SET) == (off_t) -1) {
 		errno = 0;
 		return NULL;
 	}
@@ -588,8 +558,8 @@ static struct blkid_bufinfo *read_buffer(blkid_probe pr, uint64_t real_off, uint
 	bf->off = real_off;
 	INIT_LIST_HEAD(&bf->bufs);
 
-	DBG(LOWPROBE, ul_debug("\tread %p: off=%"PRIu64" len=%"PRIu64"",
-	                       bf->data, real_off, len));
+	DBG(LOWPROBE, ul_debug("\tread: off=%"PRIu64" len=%"PRIu64"",
+	                       real_off, len));
 
 	ret = read(pr->fd, bf->data, len);
 	if (ret != (ssize_t) len) {
@@ -606,13 +576,68 @@ static struct blkid_bufinfo *read_buffer(blkid_probe pr, uint64_t real_off, uint
 	return bf;
 }
 
+/*
+ * Search in buffers we already in memory
+ */
+static struct blkid_bufinfo *get_cached_buffer(blkid_probe pr, uint64_t off, uint64_t len)
+{
+	uint64_t real_off = pr->off + off;
+	struct list_head *p;
+
+	list_for_each(p, &pr->buffers) {
+		struct blkid_bufinfo *x =
+				list_entry(p, struct blkid_bufinfo, bufs);
+
+		if (real_off >= x->off && real_off + len <= x->off + x->len) {
+			DBG(BUFFER, ul_debug("\treuse: off=%"PRIu64" len=%"PRIu64" (for off=%"PRIu64" len=%"PRIu64")",
+						x->off, x->len, real_off, len));
+			return x;
+		}
+	}
+	return NULL;
+}
+
+/*
+ * Zeroize in-memory data in already read buffer. The next blkid_probe_get_buffer()
+ * will return modified buffer. This is usable when you want to call the same probing
+ * function more than once and hide previously detected magic strings.
+ *
+ * See blkid_probe_hide_range().
+ */
+static int hide_buffer(blkid_probe pr, uint64_t off, uint64_t len)
+{
+	uint64_t real_off = pr->off + off;
+	struct list_head *p;
+	int ct = 0;
+
+	list_for_each(p, &pr->buffers) {
+		struct blkid_bufinfo *x =
+			list_entry(p, struct blkid_bufinfo, bufs);
+		unsigned char *data;
+
+		if (real_off >= x->off && real_off + len <= x->off + x->len) {
+
+			assert(x->off <= real_off);
+			assert(x->off + x->len >= real_off + len);
+
+			data = real_off ? x->data + (real_off - x->off) : x->data;
+
+			DBG(BUFFER, ul_debug("\thiding: off=%"PRIu64" len=%"PRIu64,
+						off, len));
+			memset(data, 0, len);
+			ct++;
+		}
+	}
+	return ct == 0 ? -EINVAL : 0;
+}
+
+
 /*
  * Note that @off is offset within probing area, the probing area is defined by
  * pr->off and pr->size.
  */
 unsigned char *blkid_probe_get_buffer(blkid_probe pr, uint64_t off, uint64_t len)
 {
-	struct list_head *p;
 	struct blkid_bufinfo *bf = NULL;
 	uint64_t real_off = pr->off + off;
 
@@ -647,20 +672,8 @@ unsigned char *blkid_probe_get_buffer(blkid_probe pr, uint64_t off, uint64_t len
 				pr->off + off - pr->parent->off, len);
 	}
 
-	/* try buffers we already have in memory */
-	list_for_each(p, &pr->buffers) {
-		struct blkid_bufinfo *x =
-				list_entry(p, struct blkid_bufinfo, bufs);
-
-		if (real_off >= x->off && real_off + len <= x->off + x->len) {
-			DBG(BUFFER, ul_debug("\treuse %p: off=%"PRIu64" len=%"PRIu64" (for off=%"PRIu64" len=%"PRIu64")",
-						x->data, x->off, x->len, real_off, len));
-			bf = x;
-			break;
-		}
-	}
-
-	/* not found; read from disk */
+	/* try buffers we already have in memory or read from device */
+	bf = get_cached_buffer(pr, off, len);
 	if (!bf) {
 		bf = read_buffer(pr, real_off, len);
 		if (!bf)
@@ -676,14 +689,27 @@ unsigned char *blkid_probe_get_buffer(blkid_probe pr, uint64_t off, uint64_t len
 	return real_off ? bf->data + (real_off - bf->off) : bf->data;
 }
 
-static void blkid_probe_reset_buffer(blkid_probe pr)
+/**
+ * blkid_probe_reset_buffers:
+ * @pr: prober
+ *
+ * libblkid reuse all already read buffers from the device. The buffers may be
+ * modified by blkid_probe_hide_range(). This function reset and free all
+ * cached buffers. The next blkid_do_probe() will read all data from the
+ * device.
+ *
+ * Returns: <0 in case of failure, or 0 on success.
+ */
+int blkid_probe_reset_buffers(blkid_probe pr)
 {
 	uint64_t ct = 0, len = 0;
 
-	if (list_empty(&pr->buffers))
-		return;
+	pr->flags &= ~BLKID_FL_MODIF_BUFF;
 
-	DBG(BUFFER, ul_debug("Resetting probing buffers pr=%p", pr));
+	if (list_empty(&pr->buffers))
+		return 0;
+
+	DBG(BUFFER, ul_debug("Resetting probing buffers"));
 
 	while (!list_empty(&pr->buffers)) {
 		struct blkid_bufinfo *bf = list_entry(pr->buffers.next,
@@ -692,8 +718,8 @@ static void blkid_probe_reset_buffer(blkid_probe pr)
 		len += bf->len;
 		list_del(&bf->bufs);
 
-		DBG(BUFFER, ul_debug(" remove buffer: %p [off=%"PRIu64", len=%"PRIu64"]",
-		                     bf->data, bf->off, bf->len));
+		DBG(BUFFER, ul_debug(" remove buffer: [off=%"PRIu64", len=%"PRIu64"]",
+		                     bf->off, bf->len));
 		free(bf);
 	}
 
@@ -701,6 +727,34 @@ static void blkid_probe_reset_buffer(blkid_probe pr)
 			len, ct));
 
 	INIT_LIST_HEAD(&pr->buffers);
+
+	return 0;
+}
+
+/**
+ * blkid_probe_hide_range:
+ * @pr: prober
+ * @off: start of the range
+ * @len: size of the range
+ *
+ * This function modifies in-memory cached data from the device. The specified
+ * range is zeroized. This is usable together with blkid_probe_step_back().
+ * The next blkid_do_probe() will not see specified area.
+ *
+ * Note that this is usable for already (by library) read data, and this
+ * function is not a way how to hide any large areas on your device.
+ *
+ * The function blkid_probe_reset_buffers() reverts all.
+ *
+ * Returns: <0 in case of failure, or 0 on success.
+ */
+int blkid_probe_hide_range(blkid_probe pr, uint64_t off, uint64_t len)
+{
+	int rc = hide_buffer(pr, off, len);
+
+	if (rc == 0)
+		pr->flags |= BLKID_FL_MODIF_BUFF;
+	return rc;
 }
 
 static void blkid_probe_reset_values(blkid_probe pr)
@@ -708,7 +762,7 @@ static void blkid_probe_reset_values(blkid_probe pr)
 	if (list_empty(&pr->values))
 		return;
 
-	DBG(LOWPROBE, ul_debug("resetting results pr=%p", pr));
+	DBG(LOWPROBE, ul_debug("resetting results"));
 
 	while (!list_empty(&pr->values)) {
 		struct blkid_prval *v = list_entry(pr->values.next,
@@ -742,7 +796,7 @@ static int is_sector_readable(int fd, uint64_t sector)
 	char buf[512];
 	ssize_t sz;
 
-	if (blkid_llseek(fd, sector * 512, SEEK_SET) < 0)
+	if (lseek(fd, sector * 512, SEEK_SET) == (off_t) -1)
 		goto failed;
 
 	sz = read(fd, buf, sizeof(buf));
@@ -791,10 +845,15 @@ failed:
  * @off: begin of probing area
  * @size: size of probing area (zero means whole device/file)
  *
- * Assigns the device to probe control struct, resets internal buffers and
- * resets the current probing.
+ * Assigns the device to probe control struct, resets internal buffers, resets
+ * the current probing, and close previously associated device (if open by
+ * libblkid).
  *
- * Returns: -1 in case of failure, or 0 on success.
+ * If @fd is < 0 than only resets the prober and returns 1. Note that
+ * blkid_reset_probe() keeps the device associated with the prober, but
+ * blkid_probe_set_device() does complete reset.
+ *
+ * Returns: -1 in case of failure, 0 on success and 1 on reset.
  */
 int blkid_probe_set_device(blkid_probe pr, int fd,
 		blkid_loff_t off, blkid_loff_t size)
@@ -804,11 +863,16 @@ int blkid_probe_set_device(blkid_probe pr, int fd,
 	char *dm_uuid = NULL;
 
 	blkid_reset_probe(pr);
-	blkid_probe_reset_buffer(pr);
+	blkid_probe_reset_buffers(pr);
 
 	if ((pr->flags & BLKID_FL_PRIVATE_FD) && pr->fd >= 0)
 		close(pr->fd);
 
+	if (pr->disk_probe) {
+		blkid_free_probe(pr->disk_probe);
+		pr->disk_probe = NULL;
+	}
+
 	pr->flags &= ~BLKID_FL_PRIVATE_FD;
 	pr->flags &= ~BLKID_FL_TINY_DEV;
 	pr->flags &= ~BLKID_FL_CDROM_DEV;
@@ -824,6 +888,9 @@ int blkid_probe_set_device(blkid_probe pr, int fd,
 	pr->wipe_size = 0;
 	pr->wipe_chain = NULL;
 
+	if (fd < 0)
+		return 1;
+
 #if defined(POSIX_FADV_RANDOM) && defined(HAVE_POSIX_FADVISE)
 	/* Disable read-ahead */
 	posix_fadvise(fd, 0, 0, POSIX_FADV_RANDOM);
@@ -866,8 +933,8 @@ int blkid_probe_set_device(blkid_probe pr, int fd,
 		pr->flags |= BLKID_FL_TINY_DEV;
 
 	if (S_ISBLK(sb.st_mode) &&
-	    sysfs_devno_is_lvm_private(sb.st_rdev, &dm_uuid)) {
-		DBG(LOWPROBE, ul_debug("ignore private LVM device"));
+	    sysfs_devno_is_dm_private(sb.st_rdev, &dm_uuid)) {
+		DBG(LOWPROBE, ul_debug("ignore private device mapper device"));
 		pr->flags |= BLKID_FL_NOSCAN_DEV;
 	}
 
@@ -878,6 +945,14 @@ int blkid_probe_set_device(blkid_probe pr, int fd,
 	    blkid_probe_is_wholedisk(pr) &&
 	    ioctl(fd, CDROM_GET_CAPABILITY, NULL) >= 0) {
 
+# ifdef CDROM_DRIVE_STATUS
+		switch (ioctl(fd, CDROM_DRIVE_STATUS, CDSL_CURRENT)) {
+		case CDS_TRAY_OPEN:
+		case CDS_NO_DISC:
+			errno = ENOMEDIUM;
+			goto err;
+		}
+# endif
 		pr->flags |= BLKID_FL_CDROM_DEV;
 		cdrom_size_correction(pr);
 	}
@@ -907,9 +982,9 @@ int blkid_probe_get_dimension(blkid_probe pr, uint64_t *off, uint64_t *size)
 int blkid_probe_set_dimension(blkid_probe pr, uint64_t off, uint64_t size)
 {
 	DBG(LOWPROBE, ul_debug(
-		"changing probing area pr=%p: size=%"PRIu64", off=%"PRIu64" "
+		"changing probing area: size=%"PRIu64", off=%"PRIu64" "
 		"-to-> size=%"PRIu64", off=%"PRIu64"",
-		pr, pr->size, pr->off, size, off));
+		pr->size, pr->off, size, off));
 
 	pr->off = off;
 	pr->size = size;
@@ -918,7 +993,7 @@ int blkid_probe_set_dimension(blkid_probe pr, uint64_t off, uint64_t size)
 	if (pr->size <= 1440ULL * 1024ULL && !S_ISCHR(pr->mode))
 		pr->flags |= BLKID_FL_TINY_DEV;
 
-	blkid_probe_reset_buffer(pr);
+	blkid_probe_reset_buffers(pr);
 
 	return 0;
 }
@@ -972,7 +1047,7 @@ int blkid_probe_get_idmag(blkid_probe pr, const struct blkid_idinfo *id,
 
 static inline void blkid_probe_start(blkid_probe pr)
 {
-	DBG(LOWPROBE, ul_debug("%p: start probe", pr));
+	DBG(LOWPROBE, ul_debug("start probe"));
 	pr->cur_chain = NULL;
 	pr->prob_flags = 0;
 	blkid_probe_set_wiper(pr, 0, 0);
@@ -980,7 +1055,7 @@ static inline void blkid_probe_start(blkid_probe pr)
 
 static inline void blkid_probe_end(blkid_probe pr)
 {
-	DBG(LOWPROBE, ul_debug("%p: end probe", pr));
+	DBG(LOWPROBE, ul_debug("end probe"));
 	pr->cur_chain = NULL;
 	pr->prob_flags = 0;
 	blkid_probe_set_wiper(pr, 0, 0);
@@ -1089,7 +1164,8 @@ int blkid_do_probe(blkid_probe pr)
  *
  * After successful signature removing the @pr prober will be moved one step
  * back and the next blkid_do_probe() call will again call previously called
- * probing function.
+ * probing function. All in-memory cached data from the device are always
+ * reset.
  *
  *  <example>
  *  <title>wipe all filesystems or raids from the device</title>
@@ -1105,7 +1181,7 @@ int blkid_do_probe(blkid_probe pr)
  *  </programlisting>
  * </example>
  *
- * See also blkid_probe_step_back() if you cannot use this build-in wipe
+ * See also blkid_probe_step_back() if you cannot use this built-in wipe
  * function, but you want to use libblkid probing as a source for wiping.
  *
  * Returns: 0 on success, and -1 in case of error.
@@ -1114,7 +1190,7 @@ int blkid_do_wipe(blkid_probe pr, int dryrun)
 {
 	const char *off = NULL;
 	size_t len = 0;
-	uint64_t offset, l;
+	uint64_t offset, magoff;
 	char buf[BUFSIZ];
 	int fd, rc = 0;
 	struct blkid_chain *chn;
@@ -1141,7 +1217,8 @@ int blkid_do_wipe(blkid_probe pr, int dryrun)
 	if (rc || len == 0 || off == NULL)
 		return 0;
 
-	offset = strtoumax(off, NULL, 10) + pr->off;
+	magoff = strtoumax(off, NULL, 10);
+	offset = magoff + pr->off;
 	fd = blkid_probe_get_fd(pr);
 	if (fd < 0)
 		return -1;
@@ -1153,16 +1230,25 @@ int blkid_do_wipe(blkid_probe pr, int dryrun)
 	    "do_wipe [offset=0x%"PRIx64" (%"PRIu64"), len=%zu, chain=%s, idx=%d, dryrun=%s]\n",
 	    offset, offset, len, chn->driver->name, chn->idx, dryrun ? "yes" : "not"));
 
-	l = blkid_llseek(fd, offset, SEEK_SET);
-	if ((blkid_loff_t)l == (off_t) -1)
+	if (lseek(fd, offset, SEEK_SET) == (off_t) -1)
 		return -1;
 
 	memset(buf, 0, len);
 
 	if (!dryrun && len) {
+		/* wipen on device */
 		if (write_all(fd, buf, len))
 			return -1;
 		fsync(fd);
+		pr->flags &= ~BLKID_FL_MODIF_BUFF;	/* be paranoid */
+
+		return blkid_probe_step_back(pr);
+
+	}
+
+	if (dryrun) {
+		/* wipe in memory only */
+		blkid_probe_hide_range(pr, magoff, len);
 		return blkid_probe_step_back(pr);
 	}
 
@@ -1180,6 +1266,10 @@ int blkid_do_wipe(blkid_probe pr, int dryrun)
  * This is necessary for example if you erase or modify on-disk superblock
  * according to the current libblkid probing result.
  *
+ * Note that blkid_probe_hide_range() changes semantic of this function and
+ * cached buffers are not reset, but library uses in-memory modified
+ * buffers to call the next probing function.
+ *
  * <example>
  *  <title>wipe all superblock, but use libblkid only for probing</title>
  *  <programlisting>
@@ -1226,7 +1316,8 @@ int blkid_probe_step_back(blkid_probe pr)
 	if (!chn)
 		return -1;
 
-	blkid_probe_reset_buffer(pr);
+	if (!(pr->flags & BLKID_FL_MODIF_BUFF))
+		blkid_probe_reset_buffers(pr);
 
 	if (chn->idx >= 0) {
 		chn->idx--;
@@ -1377,10 +1468,11 @@ struct blkid_prval *blkid_probe_assign_value(blkid_probe pr, const char *name)
 {
 	struct blkid_prval *v;
 
-	v = blkid_probe_new_value();
+	v = calloc(1, sizeof(struct blkid_prval));
 	if (!v)
 		return NULL;
 
+	INIT_LIST_HEAD(&v->prvals);
 	v->name = name;
 	v->chain = pr->cur_chain;
 	list_add_tail(&v->prvals, &pr->values);
@@ -1389,24 +1481,13 @@ struct blkid_prval *blkid_probe_assign_value(blkid_probe pr, const char *name)
 	return v;
 }
 
-static struct blkid_prval *blkid_probe_new_value(void)
-{
-	struct blkid_prval *v = calloc(1, sizeof(struct blkid_prval));
-	if (!v)
-		return NULL;
-
-	INIT_LIST_HEAD(&v->prvals);
-
-	return v;
-}
-
 /* Note that value data is always terminated by zero to keep things robust,
  * this extra zero is not count to the value length. It's caller responsibility
  * to set proper value length (for strings we count terminator to the length,
  * for binary data it's without terminator).
  */
 int blkid_probe_value_set_data(struct blkid_prval *v,
-		unsigned char *data, size_t len)
+		const unsigned char *data, size_t len)
 {
 	v->data = calloc(1, len + 1);	/* always terminate by \0 */
 
@@ -1418,7 +1499,7 @@ int blkid_probe_value_set_data(struct blkid_prval *v,
 }
 
 int blkid_probe_set_value(blkid_probe pr, const char *name,
-		unsigned char *data, size_t len)
+		const unsigned char *data, size_t len)
 {
 	struct blkid_prval *v;
 
@@ -1463,7 +1544,7 @@ int blkid_probe_sprintf_value(blkid_probe pr, const char *name,
 }
 
 int blkid_probe_set_magic(blkid_probe pr, uint64_t offset,
-			size_t len, unsigned char *magic)
+			size_t len, const unsigned char *magic)
 {
 	int rc = 0;
 	struct blkid_chain *chn = blkid_probe_get_chain(pr);
@@ -1824,13 +1905,6 @@ struct blkid_prval *__blkid_probe_lookup_value(blkid_probe pr, const char *name)
 
 /* converts DCE UUID (uuid[16]) to human readable string
  * - the @len should be always 37 */
-#ifdef HAVE_LIBUUID
-void blkid_unparse_uuid(const unsigned char *uuid, char *str,
-			size_t len __attribute__((__unused__)))
-{
-	uuid_unparse(uuid, str);
-}
-#else
 void blkid_unparse_uuid(const unsigned char *uuid, char *str, size_t len)
 {
 	snprintf(str, len,
@@ -1841,7 +1915,6 @@ void blkid_unparse_uuid(const unsigned char *uuid, char *str, size_t len)
 		uuid[8], uuid[9],
 		uuid[10], uuid[11], uuid[12], uuid[13], uuid[14],uuid[15]);
 }
-#endif
 
 /* like uuid_is_null() from libuuid, but works with arbitrary size of UUID */
 int blkid_uuid_is_empty(const unsigned char *buf, size_t len)
@@ -1883,8 +1956,8 @@ size_t blkid_ltrim_whitespace(unsigned char *str)
  * wiped area then the signature has been added later and LVM superblock
  * should be ignore.
  *
- * Note that this heuristic is not 100% reliable, for example "pvcreate --zero
- * n" allows to keep the begin of the device unmodified. It's probably better
+ * Note that this heuristic is not 100% reliable, for example "pvcreate --zero n"
+ * can be used to keep the begin of the device unmodified. It's probably better
  * to use this heuristic for conflicts between superblocks and partition tables
  * than for conflicts between filesystem superblocks -- existence of unwanted
  * partition table is very unusual, because PT is pretty visible (parsed and
@@ -1926,7 +1999,6 @@ void blkid_probe_set_wiper(blkid_probe pr, uint64_t off, uint64_t size)
 			chn->driver->name,
 			chn->driver->idinfos[chn->idx]->name,
 			pr->wipe_off, pr->wipe_size));
-	return;
 }
 
 /*
