@@ -1,10 +1,14 @@
+/* SPDX-License-Identifier: LGPL-2.1-or-later */
 /*
- * Copyright (C) 2009 Karel Zak <kzak@redhat.com>
+ * This file is part of libmount from util-linux project.
  *
- * This file may be redistributed under the terms of the
- * GNU Lesser General Public License.
+ * Copyright (C) 2009-2018 Karel Zak <kzak@redhat.com>
+ *
+ * libmount is free software; you can redistribute it and/or modify it
+ * under the terms of the GNU Lesser General Public License as published by
+ * the Free Software Foundation; either version 2.1 of the License, or
+ * (at your option) any later version.
  */
-
 #ifdef HAVE_SCANDIRAT
 #ifndef __USE_GNU
 #define __USE_GNU
@@ -15,205 +19,286 @@
 #include <limits.h>
 #include <dirent.h>
 #include <fcntl.h>
+#include <sys/stat.h>
 
-#include "at.h"
+#include "fileutils.h"
 #include "mangle.h"
 #include "mountP.h"
 #include "pathnames.h"
 #include "strutils.h"
 
-static inline char *skip_spaces(char *s)
-{
-	assert(s);
+struct libmnt_parser {
+	FILE	*f;		/* fstab, mtab, swaps or mountinfo ... */
+	const char *filename;	/* file name or NULL */
+	char	*buf;		/* buffer (the current line content) */
+	size_t	bufsiz;		/* size of the buffer */
+	size_t	line;		/* current line */
+};
 
-	while (*s == ' ' || *s == '\t')
-		s++;
-	return s;
+static void parser_cleanup(struct libmnt_parser *pa)
+{
+	if (!pa)
+		return;
+	free(pa->buf);
+	memset(pa, 0, sizeof(*pa));
 }
 
-static int next_number(char **s, int *num)
+static const char *next_s32(const char *s, int *num, int *rc)
 {
 	char *end = NULL;
 
-	assert(num);
-	assert(s);
+	if (!s || !*s)
+		return s;
 
-	*s = skip_spaces(*s);
-	if (!**s)
-		return -1;
-	*num = strtol(*s, &end, 10);
-	if (end == NULL || *s == end)
-	       return -1;
-
-	*s = end;
-
-	/* valid end of number is a space or a terminator */
+	*rc = -EINVAL;
+	*num = strtol(s, &end, 10);
+	if (end == NULL || s == end)
+	       return s;
 	if (*end == ' ' || *end == '\t' || *end == '\0')
-		return 0;
-	return -1;
+		*rc = 0;
+	return end;
+}
+
+static const char *next_u64(const char *s, uint64_t *num, int *rc)
+{
+	char *end = NULL;
+
+	if (!s || !*s)
+		return s;
+
+	*rc = -EINVAL;
+	*num = (uint64_t) strtoumax(s, &end, 10);
+	if (end == NULL || s == end)
+	       return s;
+	if (*end == ' ' || *end == '\t' || *end == '\0')
+		*rc = 0;
+	return end;
+}
+
+static inline const char *skip_separator(const char *p)
+{
+	while (p && (*p == ' ' || *p == '\t'))
+		++p;
+	return p;
+}
+
+static inline const char *skip_nonspearator(const char *p)
+{
+	while (p && *p && !(*p == ' ' || *p == '\t'))
+		p++;
+	return p;
 }
 
 /*
  * Parses one line from {fs,m}tab
  */
-static int mnt_parse_table_line(struct libmnt_fs *fs, char *s)
+static int mnt_parse_table_line(struct libmnt_fs *fs, const char *s)
 {
-	int rc, n = 0, xrc;
-	char *src = NULL, *fstype = NULL, *optstr = NULL;
-
-	rc = sscanf(s,	UL_SCNsA" "	/* (1) source */
-			UL_SCNsA" "	/* (2) target */
-			UL_SCNsA" "	/* (3) FS type */
-			UL_SCNsA" "	/* (4) options */
-			"%n",		/* byte count */
-
-			&src,
-			&fs->target,
-			&fstype,
-			&optstr,
-			&n);
-	xrc = rc;
-
-	if (rc == 3 || rc == 4) {			/* options are optional */
-		unmangle_string(src);
-		unmangle_string(fs->target);
-		unmangle_string(fstype);
-
-		if (optstr && *optstr)
-			unmangle_string(optstr);
-
-		/* note that __foo functions do not reallocate the string
-		 */
-		rc = __mnt_fs_set_source_ptr(fs, src);
-		if (!rc) {
-			src = NULL;
-			rc = __mnt_fs_set_fstype_ptr(fs, fstype);
-			if (!rc)
-				fstype = NULL;
-		}
-		if (!rc && optstr)
-			rc = mnt_fs_set_options(fs, optstr);
-		free(optstr);
-		optstr = NULL;
-	} else {
-		DBG(TAB, mnt_debug("tab parse error: [sscanf rc=%d]: '%s'", rc, s));
-		rc = -EINVAL;
-	}
-
-	if (rc) {
-		free(src);
-		free(fstype);
-		free(optstr);
-		DBG(TAB, mnt_debug("tab parse error: [set vars, rc=%d]\n", rc));
-		return rc;	/* error */
-	}
+	int rc = 0;
+	char *p = NULL;
 
 	fs->passno = fs->freq = 0;
 
-	if (xrc == 4 && n)
-		s = skip_spaces(s + n);
-	if (xrc == 4 && *s) {
-		if (next_number(&s, &fs->freq) != 0) {
-			if (*s) {
-				DBG(TAB, mnt_debug("tab parse error: [freq]"));
-				rc = -EINVAL;
-			}
-		} else if (next_number(&s, &fs->passno) != 0 && *s) {
-			DBG(TAB, mnt_debug("tab parse error: [passno]"));
-			rc = -EINVAL;
-		}
+	/* (1) source */
+	p = unmangle(s, &s);
+	if (!p || (rc = __mnt_fs_set_source_ptr(fs, p))) {
+		DBG(TAB, ul_debug("tab parse error: [source]"));
+		free(p);
+		goto fail;
 	}
 
+	s = skip_separator(s);
+
+	/* (2) target */
+	fs->target = unmangle(s, &s);
+	if (!fs->target) {
+		DBG(TAB, ul_debug("tab parse error: [target]"));
+		goto fail;
+	}
+
+	s = skip_separator(s);
+
+	/* (3) FS type */
+	p = unmangle(s, &s);
+	if (!p || (rc = __mnt_fs_set_fstype_ptr(fs, p))) {
+		DBG(TAB, ul_debug("tab parse error: [fstype]"));
+		free(p);
+		goto fail;
+	}
+
+	s = skip_separator(s);
+
+	/* (4) options (optional) */
+	p = unmangle(s, &s);
+	if (p && (rc = mnt_fs_set_options(fs, p))) {
+		DBG(TAB, ul_debug("tab parse error: [options]"));
+		free(p);
+		goto fail;
+	}
+	if (!p)
+		goto done;
+	free(p);
+
+	s = skip_separator(s);
+	if (!s || !*s)
+		goto done;
+
+	/* (5) freq (optional) */
+	s = next_s32(s, &fs->freq, &rc);
+	if (s && *s && rc) {
+		DBG(TAB, ul_debug("tab parse error: [freq]"));
+		goto fail;
+	}
+
+	s = skip_separator(s);
+	if (!s || !*s)
+		goto done;
+
+	/* (6) passno (optional) */
+	s = next_s32(s, &fs->passno, &rc);
+	if (s && *s && rc) {
+		DBG(TAB, ul_debug("tab parse error: [passno]"));
+		goto fail;
+	}
+
+done:
+	return 0;
+fail:
+	if (rc == 0)
+		rc = -EINVAL;
+	DBG(TAB, ul_debug("tab parse error on: '%s' [rc=%d]", s, rc));
 	return rc;
 }
 
+
 /*
  * Parses one line from a mountinfo file
  */
-static int mnt_parse_mountinfo_line(struct libmnt_fs *fs, char *s)
+static int mnt_parse_mountinfo_line(struct libmnt_fs *fs, const char *s)
 {
-	int rc, end = 0;
+	int rc = 0;
 	unsigned int maj, min;
-	char *fstype = NULL, *src = NULL, *p;
+	char *p;
 
-	rc = sscanf(s,	"%u "		/* (1) id */
-			"%u "		/* (2) parent */
-			"%u:%u "	/* (3) maj:min */
-			UL_SCNsA" "	/* (4) mountroot */
-			UL_SCNsA" "	/* (5) target */
-			UL_SCNsA	/* (6) vfs options (fs-independent) */
-			"%n",		/* number of read bytes */
+	fs->flags |= MNT_FS_KERNEL;
 
-			&fs->id,
-			&fs->parent,
-			&maj, &min,
-			&fs->root,
-			&fs->target,
-			&fs->vfs_optstr,
-			&end);
+	/* (1) id */
+	s = next_s32(s, &fs->id, &rc);
+	if (!s || !*s || rc) {
+		DBG(TAB, ul_debug("tab parse error: [id]"));
+		goto fail;
+	}
 
-	if (rc >= 7 && end > 0)
-		s += end;
+	s = skip_separator(s);
+
+	/* (2) parent */
+	s = next_s32(s, &fs->parent, &rc);
+	if (!s || !*s || rc) {
+		DBG(TAB, ul_debug("tab parse error: [parent]"));
+		goto fail;
+	}
+
+	s = skip_separator(s);
+
+	/* (3) maj:min */
+	if (sscanf(s, "%u:%u", &maj, &min) != 2) {
+		DBG(TAB, ul_debug("tab parse error: [maj:min]"));
+		goto fail;
+	}
+	fs->devno = makedev(maj, min);
+	s = skip_nonspearator(s);
+	s = skip_separator(s);
+
+	/* (4) mountroot */
+	fs->root = unmangle(s, &s);
+	if (!fs->root) {
+		DBG(TAB, ul_debug("tab parse error: [mountroot]"));
+		goto fail;
+	}
+
+	s = skip_separator(s);
+
+	/* (5) target */
+	fs->target = unmangle(s, &s);
+	if (!fs->target) {
+		DBG(TAB, ul_debug("tab parse error: [target]"));
+		goto fail;
+	}
+
+	/* remove "\040(deleted)" suffix */
+	p = (char *) endswith(fs->target, PATH_DELETED_SUFFIX);
+	if (p && *p)
+		*p = '\0';
+
+	s = skip_separator(s);
+
+	/* (6) vfs options (fs-independent) */
+	fs->vfs_optstr = unmangle(s, &s);
+	if (!fs->vfs_optstr) {
+		DBG(TAB, ul_debug("tab parse error: [VFS options]"));
+		goto fail;
+	}
 
 	/* (7) optional fields, terminated by " - " */
 	p = strstr(s, " - ");
 	if (!p) {
-		DBG(TAB, mnt_debug("mountinfo parse error: separator not found"));
+		DBG(TAB, ul_debug("mountinfo parse error: separator not found"));
 		return -EINVAL;
 	}
 	if (p > s + 1)
 		fs->opt_fields = strndup(s + 1, p - s - 1);
-	s = p + 3;
 
-	rc += sscanf(s,	UL_SCNsA" "	/* (8) FS type */
-			UL_SCNsA" "	/* (9) source */
-			UL_SCNsA,	/* (10) fs options (fs specific) */
+	s = skip_separator(p + 3);
 
-			&fstype,
-			&src,
-			&fs->fs_optstr);
-
-	if (rc >= 10) {
-		size_t sz;
-
-		fs->flags |= MNT_FS_KERNEL;
-		fs->devno = makedev(maj, min);
-
-		/* remove "(deleted)" suffix */
-		sz = strlen(fs->target);
-		if (sz > PATH_DELETED_SUFFIX_SZ) {
-			char *p = fs->target + (sz - PATH_DELETED_SUFFIX_SZ);
-
-			if (strcmp(p, PATH_DELETED_SUFFIX) == 0)
-				*p = '\0';
-		}
-
-		unmangle_string(fs->root);
-		unmangle_string(fs->target);
-		unmangle_string(fs->vfs_optstr);
-		unmangle_string(fstype);
-		unmangle_string(src);
-		unmangle_string(fs->fs_optstr);
-
-		rc = __mnt_fs_set_fstype_ptr(fs, fstype);
-		if (!rc) {
-			fstype = NULL;
-			rc = __mnt_fs_set_source_ptr(fs, src);
-			if (!rc)
-				src = NULL;
-		}
-
-		/* merge VFS and FS options to one string */
-		fs->optstr = mnt_fs_strdup_options(fs);
-		if (!fs->optstr)
-			rc = -ENOMEM;
-	} else {
-		free(fstype);
-		free(src);
-		DBG(TAB, mnt_debug(
-			"mountinfo parse error [sscanf rc=%d]: '%s'", rc, s));
-		rc = -EINVAL;
+	/* (8) FS type */
+	p = unmangle(s, &s);
+	if (!p || (rc = __mnt_fs_set_fstype_ptr(fs, p))) {
+		DBG(TAB, ul_debug("tab parse error: [fstype]"));
+		free(p);
+		goto fail;
 	}
+
+	/* (9) source -- maybe empty string */
+	if (!s || !*s) {
+		DBG(TAB, ul_debug("tab parse error: [source]"));
+		goto fail;
+	} else if (*s == ' ' && *(s+1) == ' ') {
+		if ((rc = mnt_fs_set_source(fs, ""))) {
+			DBG(TAB, ul_debug("tab parse error: [empty source]"));
+			goto fail;
+		}
+	} else {
+		s = skip_separator(s);
+		p = unmangle(s, &s);
+		if (!p || (rc = __mnt_fs_set_source_ptr(fs, p))) {
+			DBG(TAB, ul_debug("tab parse error: [regular source]"));
+			free(p);
+			goto fail;
+		}
+	}
+
+	s = skip_separator(s);
+
+	/* (10) fs options (fs specific) */
+	fs->fs_optstr = unmangle(s, &s);
+	if (!fs->fs_optstr) {
+		DBG(TAB, ul_debug("tab parse error: [FS options]"));
+		goto fail;
+	}
+
+	/* merge VFS and FS options to one string */
+	fs->optstr = mnt_fs_strdup_options(fs);
+	if (!fs->optstr) {
+		rc = -ENOMEM;
+		DBG(TAB, ul_debug("tab parse error: [merge VFS and FS options]"));
+		goto fail;
+	}
+
+	return 0;
+fail:
+	if (rc == 0)
+		rc = -EINVAL;
+	DBG(TAB, ul_debug("tab parse error on: '%s' [rc=%d]", s, rc));
 	return rc;
 }
 
@@ -230,7 +315,7 @@ static int mnt_parse_utab_line(struct libmnt_fs *fs, const char *s)
 	assert(!fs->target);
 
 	while (p && *p) {
-		char *end = NULL;
+		const char *end = NULL;
 
 		while (*p == ' ') p++;
 		if (!*p)
@@ -240,7 +325,8 @@ static int mnt_parse_utab_line(struct libmnt_fs *fs, const char *s)
 			char *v = unmangle(p + 4, &end);
 			if (!v)
 				goto enomem;
-			__mnt_fs_set_source_ptr(fs, v);
+			if (__mnt_fs_set_source_ptr(fs, v))
+				free(v);
 
 		} else if (!fs->target && !strncmp(p, "TARGET=", 7)) {
 			fs->target = unmangle(p + 7, &end);
@@ -277,56 +363,76 @@ static int mnt_parse_utab_line(struct libmnt_fs *fs, const char *s)
 
 	return 0;
 enomem:
-	DBG(TAB, mnt_debug("utab parse error: ENOMEM"));
+	DBG(TAB, ul_debug("utab parse error: ENOMEM"));
 	return -ENOMEM;
 }
 
 /*
  * Parses one line from /proc/swaps
  */
-static int mnt_parse_swaps_line(struct libmnt_fs *fs, char *s)
+static int mnt_parse_swaps_line(struct libmnt_fs *fs, const char *s)
 {
-	uintmax_t fsz, usz;
-	int rc;
-	char *src = NULL;
+	uint64_t num;
+	int rc = 0;
+	char *p;
 
-	rc = sscanf(s,	UL_SCNsA" "	/* (1) source */
-			UL_SCNsA" "	/* (2) type */
-			"%jd"		/* (3) size */
-			"%jd"		/* (4) used */
-			"%d",		/* priority */
-
-			&src,
-			&fs->swaptype,
-			&fsz,
-			&usz,
-			&fs->priority);
-
-	if (rc == 5) {
-		size_t sz;
-
-		fs->size = fsz;
-		fs->usedsize = usz;
-
-		unmangle_string(src);
-
-		/* remove "(deleted)" suffix */
-		sz = strlen(src);
-		if (sz > PATH_DELETED_SUFFIX_SZ) {
-			char *p = src + (sz - PATH_DELETED_SUFFIX_SZ);
-			if (strcmp(p, PATH_DELETED_SUFFIX) == 0)
-				*p = '\0';
-		}
-
-		rc = mnt_fs_set_source(fs, src);
-		if (!rc)
-			mnt_fs_set_fstype(fs, "swap");
-		free(src);
-	} else {
-		DBG(TAB, mnt_debug("tab parse error: [sscanf rc=%d]: '%s'", rc, s));
-		rc = -EINVAL;
+	/* (1) source */
+	p = unmangle(s, &s);
+	if (p) {
+		char *x = (char *) endswith(p, PATH_DELETED_SUFFIX);
+		if (x && *x)
+			*x = '\0';
+	}
+	if (!p || (rc = __mnt_fs_set_source_ptr(fs, p))) {
+		DBG(TAB, ul_debug("tab parse error: [source]"));
+		free(p);
+		goto fail;
 	}
 
+	s = skip_separator(s);
+
+	/* (2) type */
+	fs->swaptype = unmangle(s, &s);
+	if (!fs->swaptype) {
+		DBG(TAB, ul_debug("tab parse error: [swaptype]"));
+		goto fail;
+	}
+
+	s = skip_separator(s);
+
+	/* (3) size */
+	s = next_u64(s, &num, &rc);
+	if (!s || !*s || rc) {
+		DBG(TAB, ul_debug("tab parse error: [size]"));
+		goto fail;
+	}
+	fs->size = num;
+
+	s = skip_separator(s);
+
+	/* (4) size */
+	s = next_u64(s, &num, &rc);
+	if (!s || !*s || rc) {
+		DBG(TAB, ul_debug("tab parse error: [used size]"));
+		goto fail;
+	}
+	fs->usedsize = num;
+
+	s = skip_separator(s);
+
+	/* (5) priority */
+	s = next_s32(s, &fs->priority, &rc);
+	if (rc) {
+		DBG(TAB, ul_debug("tab parse error: [priority]"));
+		goto fail;
+	}
+
+	mnt_fs_set_fstype(fs, "swap");
+	return 0;
+fail:
+	if (rc == 0)
+		rc = -EINVAL;
+	DBG(TAB, ul_debug("tab parse error on: '%s' [rc=%d]", s, rc));
 	return rc;
 }
 
@@ -340,11 +446,11 @@ static int mnt_parse_swaps_line(struct libmnt_fs *fs, char *s)
  *
  * mountinfo: "<number> <number> ... "
  */
-static int guess_table_format(char *line)
+static int guess_table_format(const char *line)
 {
 	unsigned int a, b;
 
-	DBG(TAB, mnt_debug("trying to guess table type"));
+	DBG(TAB, ul_debug("trying to guess table type"));
 
 	if (sscanf(line, "%u %u", &a, &b) == 2)
 		return MNT_FMT_MOUNTINFO;
@@ -355,9 +461,9 @@ static int guess_table_format(char *line)
 	return MNT_FMT_FSTAB;		/* fstab, mtab or /proc/mounts */
 }
 
-static int is_comment_line(char *line)
+static int is_comment_line(const char *line)
 {
-	char *p	= skip_spaces(line);
+	const char *p = skip_blank(line);
 
 	if (p && (*p == '#' || *p == '\n'))
 		return 1;
@@ -387,16 +493,15 @@ static int is_terminated_by_blank(const char *str)
  *         1 if the line is not a comment
  *        <0 on error
  */
-static int next_comment_line(char *buf, size_t bufsz,
-			     FILE *f, char **last, int *nlines)
+static int next_comment_line(struct libmnt_parser *pa, char **last)
 {
-	if (fgets(buf, bufsz, f) == NULL)
-		return feof(f) ? 1 : -EINVAL;
+	if (getline(&pa->buf, &pa->bufsiz, pa->f) < 0)
+		return feof(pa->f) ? 1 : -errno;
 
-	++*nlines;
-	*last = strchr(buf, '\n');
+	pa->line++;
+	*last = strchr(pa->buf, '\n');
 
-	return is_comment_line(buf) ? 0 : 1;
+	return is_comment_line(pa->buf) ? 0 : 1;
 }
 
 static int append_comment(struct libmnt_table *tb,
@@ -409,7 +514,7 @@ static int append_comment(struct libmnt_table *tb,
 	if (intro && is_terminated_by_blank(mnt_table_get_intro_comment(tb)))
 		intro = 0;
 
-	DBG(TAB, mnt_debug_h(tb, "appending %s comment",
+	DBG(TAB, ul_debugobj(tb, "appending %s comment",
 			intro ? "intro" :
 			eof ? "trailing" : "fs"));
 	if (intro)
@@ -429,36 +534,35 @@ static int append_comment(struct libmnt_table *tb,
 /*
  * Read and parse the next line from {fs,m}tab or mountinfo
  */
-static int mnt_table_parse_next(struct libmnt_table *tb, FILE *f,
-				struct libmnt_fs *fs,
-				const char *filename, int *nlines)
+static int mnt_table_parse_next(struct libmnt_parser *pa,
+				struct libmnt_table *tb,
+				struct libmnt_fs *fs)
 {
-	char buf[BUFSIZ];
 	char *s;
 	int rc;
 
 	assert(tb);
-	assert(f);
+	assert(pa);
 	assert(fs);
 
 	/* read the next non-blank non-comment line */
 next_line:
 	do {
-		if (fgets(buf, sizeof(buf), f) == NULL)
+		if (getline(&pa->buf, &pa->bufsiz, pa->f) < 0)
 			return -EINVAL;
-		++*nlines;
-		s = strchr (buf, '\n');
+		pa->line++;
+		s = strchr(pa->buf, '\n');
 		if (!s) {
 			/* Missing final newline?  Otherwise an extremely */
 			/* long line - assume file was corrupted */
-			if (feof(f)) {
-				DBG(TAB, mnt_debug_h(tb,
-					"%s: no final newline",	filename));
-				s = strchr (buf, '\0');
+			if (feof(pa->f)) {
+				DBG(TAB, ul_debugobj(tb,
+					"%s: no final newline",	pa->filename));
+				s = strchr(pa->buf, '\0');
 			} else {
-				DBG(TAB, mnt_debug_h(tb,
-					"%s:%d: missing newline at line",
-					filename, *nlines));
+				DBG(TAB, ul_debugobj(tb,
+					"%s:%zu: missing newline at line",
+					pa->filename, pa->line));
 				goto err;
 			}
 		}
@@ -466,16 +570,14 @@ next_line:
 		/* comments parser */
 		if (tb->comms
 		    && (tb->fmt == MNT_FMT_GUESS || tb->fmt == MNT_FMT_FSTAB)
-		    && is_comment_line(buf)) {
+		    && is_comment_line(pa->buf)) {
 			do {
-				rc = append_comment(tb, fs, buf, feof(f));
+				rc = append_comment(tb, fs, pa->buf, feof(pa->f));
 				if (!rc)
-					rc = next_comment_line(buf,
-							sizeof(buf),
-							f, &s, nlines);
+					rc = next_comment_line(pa, &s);
 			} while (rc == 0);
 
-			if (rc == 1 && feof(f))
+			if (rc == 1 && feof(pa->f))
 				rc = append_comment(tb, fs, NULL, 1);
 			if (rc < 0)
 				return rc;
@@ -483,9 +585,9 @@ next_line:
 		}
 
 		*s = '\0';
-		if (--s >= buf && *s == '\r')
+		if (--s >= pa->buf && *s == '\r')
 			*s = '\0';
-		s = skip_spaces(buf);
+		s = (char *) skip_blank(pa->buf);
 	} while (*s == '\0' || *s == '#');
 
 	if (tb->fmt == MNT_FMT_GUESS) {
@@ -517,7 +619,7 @@ next_line:
 	if (rc == 0)
 		return 0;
 err:
-	DBG(TAB, mnt_debug_h(tb, "%s:%d: %s parse error", filename, *nlines,
+	DBG(TAB, ul_debugobj(tb, "%s:%zu: %s parse error", pa->filename, pa->line,
 				tb->fmt == MNT_FMT_MOUNTINFO ? "mountinfo" :
 				tb->fmt == MNT_FMT_SWAPS ? "swaps" :
 				tb->fmt == MNT_FMT_FSTAB ? "tab" : "utab"));
@@ -525,7 +627,7 @@ err:
 	/* by default all errors are recoverable, otherwise behavior depends on
 	 * the errcb() function. See mnt_table_set_parser_errcb().
 	 */
-	return tb->errcb ? tb->errcb(tb, filename, *nlines) : 1;
+	return tb->errcb ? tb->errcb(tb, pa->filename, pa->line) : 1;
 }
 
 static pid_t path_to_tid(const char *filename)
@@ -551,7 +653,7 @@ static pid_t path_to_tid(const char *filename)
 		tid = 0;
 		goto done;
 	}
-	DBG(TAB, mnt_debug("TID for %s is %d", filename, tid));
+	DBG(TAB, ul_debug("TID for %s is %d", filename, tid));
 done:
 	free(path);
 	return tid;
@@ -576,24 +678,105 @@ static int kernel_fs_postparse(struct libmnt_table *tb,
 	 * Convert obscure /dev/root to something more usable
 	 */
 	if (src && strcmp(src, "/dev/root") == 0) {
-		char *spec = mnt_get_kernel_cmdline_option("root=");
 		char *real = NULL;
 
-		DBG(TAB, mnt_debug_h(tb, "root FS: %s", spec));
-		if (spec)
-			real = mnt_resolve_spec(spec, tb->cache);
-		if (real) {
-			DBG(TAB, mnt_debug_h(tb, "canonical root FS: %s", real));
-			rc = mnt_fs_set_source(fs, real);
-			if (!tb->cache)
-				free(real);
+		rc = mnt_guess_system_root(fs->devno, tb->cache, &real);
+		if (rc < 0)
+			return rc;
+
+		if (rc == 0 && real) {
+			DBG(TAB, ul_debugobj(tb, "canonical root FS: %s", real));
+			rc = __mnt_fs_set_source_ptr(fs, real);
+
+		} else if (rc == 1) {
+			/* mnt_guess_system_root() returns 1 if not able to convert to
+			 * the real devname; ignore this problem */
+			rc = 0;
 		}
-		free(spec);
 	}
 
 	return rc;
 }
 
+static int __table_parse_stream(struct libmnt_table *tb, FILE *f, const char *filename)
+{
+	int rc = -1;
+	int flags = 0;
+	pid_t tid = -1;
+	struct libmnt_parser pa = { .line = 0 };
+
+	assert(tb);
+	assert(f);
+	assert(filename);
+
+	DBG(TAB, ul_debugobj(tb, "%s: start parsing [entries=%d, filter=%s]",
+				filename, mnt_table_get_nents(tb),
+				tb->fltrcb ? "yes" : "not"));
+
+	pa.filename = filename;
+	pa.f = f;
+
+	/* necessary for /proc/mounts only, the /proc/self/mountinfo
+	 * parser sets the flag properly
+	 */
+	if (filename && strcmp(filename, _PATH_PROC_MOUNTS) == 0)
+		flags = MNT_FS_KERNEL;
+
+	do {
+		struct libmnt_fs *fs;
+
+		if (feof(f)) {
+			DBG(TAB, ul_debugobj(tb, "end-of-file"));
+			break;
+		}
+		fs = mnt_new_fs();
+		if (!fs)
+			goto err;
+
+		/* parse */
+		rc = mnt_table_parse_next(&pa, tb, fs);
+
+		if (rc == 0 && tb->fltrcb && tb->fltrcb(fs, tb->fltrcb_data))
+			rc = 1;	/* filtered out by callback... */
+
+		/* add to the table */
+		if (rc == 0) {
+			rc = mnt_table_add_fs(tb, fs);
+			fs->flags |= flags;
+
+			if (rc == 0 && tb->fmt == MNT_FMT_MOUNTINFO) {
+				rc = kernel_fs_postparse(tb, fs, &tid, filename);
+				if (rc)
+					mnt_table_remove_fs(tb, fs);
+			}
+		}
+
+		/* remove reference (or deallocate on error) */
+		mnt_unref_fs(fs);
+
+		/* recoverable error */
+		if (rc > 0) {
+			DBG(TAB, ul_debugobj(tb, "recoverable error (continue)"));
+			continue;
+		}
+
+		/* fatal errors */
+		if (rc < 0 && !feof(f)) {
+			DBG(TAB, ul_debugobj(tb, "fatal error"));
+			goto err;
+		}
+	} while (1);
+
+	DBG(TAB, ul_debugobj(tb, "%s: stop parsing (%d entries)",
+				filename, mnt_table_get_nents(tb)));
+	parser_cleanup(&pa);
+	return 0;
+err:
+	DBG(TAB, ul_debugobj(tb, "%s: parse error (rc=%d)", filename, rc));
+	parser_cleanup(&pa);
+	return rc;
+}
+
 /**
  * mnt_table_parse_stream:
  * @tb: tab pointer
@@ -604,59 +787,27 @@ static int kernel_fs_postparse(struct libmnt_table *tb,
  */
 int mnt_table_parse_stream(struct libmnt_table *tb, FILE *f, const char *filename)
 {
-	int nlines = 0;
-	int rc = -1;
-	int flags = 0;
-	pid_t tid = -1;
+	int fd, rc;
+	FILE *memf = NULL;
+	char *membuf = NULL;
 
-	assert(tb);
-	assert(f);
-	assert(filename);
-
-	DBG(TAB, mnt_debug_h(tb, "%s: start parsing [entries=%d, filter=%s]",
-				filename, mnt_table_get_nents(tb),
-				tb->fltrcb ? "yes" : "not"));
-
-	/* necessary for /proc/mounts only, the /proc/self/mountinfo
-	 * parser sets the flag properly
+	/*
+	 * For /proc/#/{mountinfo,mount} we read all file to memory and use it
+	 * as memory stream. For more details see mnt_read_procfs_file().
 	 */
-	if (filename && strcmp(filename, _PATH_PROC_MOUNTS) == 0)
-		flags = MNT_FS_KERNEL;
+	if ((fd = fileno(f)) >= 0
+	    && (tb->fmt == MNT_FMT_GUESS ||
+		tb->fmt == MNT_FMT_MOUNTINFO ||
+		tb->fmt == MNT_FMT_MTAB)
+	    && is_procfs_fd(fd)
+	    && (memf = mnt_get_procfs_memstream(fd, &membuf))) {
 
-	while (!feof(f)) {
-		struct libmnt_fs *fs = mnt_new_fs();
+		rc = __table_parse_stream(tb, memf, filename);
+		fclose(memf);
+		free(membuf);
+	} else
+		rc = __table_parse_stream(tb, f, filename);
 
-		if (!fs)
-			goto err;
-
-		rc = mnt_table_parse_next(tb, f, fs, filename, &nlines);
-
-		if (!rc && tb->fltrcb && tb->fltrcb(fs, tb->fltrcb_data))
-			rc = 1;	/* filtered out by callback... */
-
-		if (!rc) {
-			rc = mnt_table_add_fs(tb, fs);
-			fs->flags |= flags;
-
-			if (rc == 0 && tb->fmt == MNT_FMT_MOUNTINFO)
-				rc = kernel_fs_postparse(tb, fs, &tid, filename);
-		}
-		mnt_unref_fs(fs);
-
-		if (rc) {
-			if (rc == 1)
-				continue;	/* recoverable error */
-			if (feof(f))
-				break;
-			goto err;		/* fatal error */
-		}
-	}
-
-	DBG(TAB, mnt_debug_h(tb, "%s: stop parsing (%d entries)",
-				filename, mnt_table_get_nents(tb)));
-	return 0;
-err:
-	DBG(TAB, mnt_debug_h(tb, "%s: parse error (rc=%d)", filename, rc));
 	return rc;
 }
 
@@ -665,7 +816,7 @@ err:
  * @tb: tab pointer
  * @filename: file
  *
- * Parses the whole table (e.g. /etc/mtab) and appends new records to the @tab.
+ * Parses the whole table (e.g. /etc/fstab) and appends new records to the @tab.
  *
  * The libmount parser ignores broken (syntax error) lines, these lines are
  * reported to the caller by the errcb() function (see mnt_table_set_parser_errcb()).
@@ -675,21 +826,50 @@ err:
 int mnt_table_parse_file(struct libmnt_table *tb, const char *filename)
 {
 	FILE *f;
-	int rc;
-
-	assert(tb);
-	assert(filename);
+	int rc, fd = -1;
 
 	if (!filename || !tb)
 		return -EINVAL;
 
-	f = fopen(filename, "r" UL_CLOEXECSTR);
+	/*
+	 * Try to use read()+poll() to realiably read all
+	 * /proc/#/{mount,mountinfo} file to memory
+	 */
+	if (tb->fmt != MNT_FMT_SWAPS
+	    && strncmp(filename, "/proc/", 6) == 0) {
+
+		FILE *memf;
+		char *membuf = NULL;
+
+		fd = open(filename, O_RDONLY|O_CLOEXEC);
+		if (fd < 0) {
+			rc = -errno;
+			goto done;
+		}
+		memf = mnt_get_procfs_memstream(fd, &membuf);
+		if (memf) {
+			rc = __table_parse_stream(tb, memf, filename);
+
+			fclose(memf);
+			free(membuf);
+			close(fd);
+			goto done;
+		}
+		/* else fallback to fopen/fdopen() */
+	}
+
+	if (fd >= 0)
+		f = fdopen(fd, "r" UL_CLOEXECSTR);
+	else
+		f = fopen(filename, "r" UL_CLOEXECSTR);
+
 	if (f) {
-		rc = mnt_table_parse_stream(tb, f, filename);
+		rc = __table_parse_stream(tb, f, filename);
 		fclose(f);
 	} else
-		return -errno;
-
+		rc = -errno;
+done:
+	DBG(TAB, ul_debugobj(tb, "parsing done [filename=%s, rc=%d]", filename, rc));
 	return rc;
 }
 
@@ -710,7 +890,7 @@ static int mnt_table_parse_dir_filter(const struct dirent *d)
 	namesz = strlen(d->d_name);
 	if (!namesz || namesz < MNT_MNTTABDIR_EXTSIZ + 1 ||
 	    strcmp(d->d_name + (namesz - MNT_MNTTABDIR_EXTSIZ),
-		   MNT_MNTTABDIR_EXT))
+		   MNT_MNTTABDIR_EXT) != 0)
 		return 0;
 
 	/* Accept this */
@@ -739,13 +919,13 @@ static int __mnt_table_parse_dir(struct libmnt_table *tb, const char *dirname)
 		struct stat st;
 		FILE *f;
 
-		if (fstat_at(dd, ".", d->d_name, &st, 0) ||
+		if (fstatat(dd, d->d_name, &st, 0) ||
 		    !S_ISREG(st.st_mode))
 			continue;
 
-		f = fopen_at(dd, ".", d->d_name, O_RDONLY|O_CLOEXEC, "r" UL_CLOEXECSTR);
+		f = fopen_at(dd, d->d_name, O_RDONLY|O_CLOEXEC, "r" UL_CLOEXECSTR);
 		if (f) {
-			mnt_table_parse_stream(tb, f, d->d_name);
+			__table_parse_stream(tb, f, d->d_name);
 			fclose(f);
 		}
 	}
@@ -779,14 +959,14 @@ static int __mnt_table_parse_dir(struct libmnt_table *tb, const char *dirname)
 		struct stat st;
 		FILE *f;
 
-		if (fstat_at(dirfd(dir), _PATH_MNTTAB_DIR, d->d_name, &st, 0) ||
+		if (fstatat(dirfd(dir), d->d_name, &st, 0) ||
 		    !S_ISREG(st.st_mode))
 			continue;
 
-		f = fopen_at(dirfd(dir), _PATH_MNTTAB_DIR, d->d_name,
+		f = fopen_at(dirfd(dir), d->d_name,
 				O_RDONLY|O_CLOEXEC, "r" UL_CLOEXECSTR);
 		if (f) {
-			mnt_table_parse_stream(tb, f, d->d_name);
+			__table_parse_stream(tb, f, d->d_name);
 			fclose(f);
 		}
 	}
@@ -818,18 +998,19 @@ int mnt_table_parse_dir(struct libmnt_table *tb, const char *dirname)
 	return __mnt_table_parse_dir(tb, dirname);
 }
 
-struct libmnt_table *__mnt_new_table_from_file(const char *filename, int fmt)
+struct libmnt_table *__mnt_new_table_from_file(const char *filename, int fmt, int empty_for_enoent)
 {
 	struct libmnt_table *tb;
 	struct stat st;
 
-	assert(filename);
 	if (!filename)
 		return NULL;
 	if (stat(filename, &st))
-		return NULL;
+		return empty_for_enoent ? mnt_new_table() : NULL;
+
 	tb = mnt_new_table();
 	if (tb) {
+		DBG(TAB, ul_debugobj(tb, "new tab for file: %s", filename));
 		tb->fmt = fmt;
 		if (mnt_table_parse_file(tb, filename) != 0) {
 			mnt_unref_table(tb);
@@ -852,8 +1033,10 @@ struct libmnt_table *__mnt_new_table_from_file(const char *filename, int fmt)
  */
 struct libmnt_table *mnt_new_table_from_file(const char *filename)
 {
-	assert(filename);
-	return __mnt_new_table_from_file(filename, MNT_FMT_GUESS);
+	if (!filename)
+		return NULL;
+
+	return __mnt_new_table_from_file(filename, MNT_FMT_GUESS, 0);
 }
 
 /**
@@ -866,7 +1049,6 @@ struct libmnt_table *mnt_new_table_from_dir(const char *dirname)
 {
 	struct libmnt_table *tb;
 
-	assert(dirname);
 	if (!dirname)
 		return NULL;
 	tb = mnt_new_table();
@@ -896,7 +1078,6 @@ struct libmnt_table *mnt_new_table_from_dir(const char *dirname)
 int mnt_table_set_parser_errcb(struct libmnt_table *tb,
 		int (*cb)(struct libmnt_table *tb, const char *filename, int line))
 {
-	assert(tb);
 	if (!tb)
 		return -EINVAL;
 	tb->errcb = cb;
@@ -911,11 +1092,10 @@ int mnt_table_set_parser_fltrcb(struct libmnt_table *tb,
 		int (*cb)(struct libmnt_fs *, void *),
 		void *data)
 {
-	assert(tb);
 	if (!tb)
 		return -EINVAL;
 
-	DBG(TAB, mnt_debug_h(tb, "%s table parser filter", cb ? "set" : "unset"));
+	DBG(TAB, ul_debugobj(tb, "%s table parser filter", cb ? "set" : "unset"));
 	tb->fltrcb = cb;
 	tb->fltrcb_data = data;
 	return 0;
@@ -934,8 +1114,6 @@ int mnt_table_set_parser_fltrcb(struct libmnt_table *tb,
  */
 int mnt_table_parse_swaps(struct libmnt_table *tb, const char *filename)
 {
-	assert(tb);
-
 	if (!tb)
 		return -EINVAL;
 	if (!filename) {
@@ -966,14 +1144,14 @@ int mnt_table_parse_fstab(struct libmnt_table *tb, const char *filename)
 	struct stat st;
 	int rc = 0;
 
-	assert(tb);
 	if (!tb)
 		return -EINVAL;
 	if (!filename)
 		filename = mnt_get_fstab_path();
-
-	if (!filename || stat(filename, &st))
+	if (!filename)
 		return -EINVAL;
+	if (stat(filename, &st) != 0)
+		return -errno;
 
 	tb->fmt = MNT_FMT_FSTAB;
 
@@ -1002,12 +1180,10 @@ static struct libmnt_fs *mnt_table_merge_user_fs(struct libmnt_table *tb, struct
 	struct libmnt_iter itr;
 	const char *optstr, *src, *target, *root, *attrs;
 
-	assert(tb);
-	assert(uf);
 	if (!tb || !uf)
 		return NULL;
 
-	DBG(TAB, mnt_debug_h(tb, "merging user fs"));
+	DBG(TAB, ul_debugobj(tb, "merging user fs"));
 
 	src = mnt_fs_get_srcpath(uf);
 	target = mnt_fs_get_target(uf);
@@ -1033,77 +1209,101 @@ static struct libmnt_fs *mnt_table_merge_user_fs(struct libmnt_table *tb, struct
 	}
 
 	if (fs) {
-		DBG(TAB, mnt_debug_h(tb, "found fs -- appending user optstr"));
+		DBG(TAB, ul_debugobj(tb, "found fs -- appending user optstr"));
 		mnt_fs_append_options(fs, optstr);
 		mnt_fs_append_attributes(fs, attrs);
 		mnt_fs_set_bindsrc(fs, mnt_fs_get_bindsrc(uf));
 		fs->flags |= MNT_FS_MERGED;
 
-		DBG(TAB, mnt_debug_h(tb, "found fs:"));
+		DBG(TAB, ul_debugobj(tb, "found fs:"));
 		DBG(TAB, mnt_fs_print_debug(fs, stderr));
 	}
 	return fs;
 }
 
-/**
- * mnt_table_parse_mtab:
- * @tb: table
- * @filename: overwrites default (/etc/mtab or $LIBMOUNT_MTAB) or NULL
- *
- * This function parses /etc/mtab or /proc/self/mountinfo +
- * /run/mount/utabs or /proc/mounts.
- *
- * See also mnt_table_set_parser_errcb().
- *
- * Returns: 0 on success or negative number in case of error.
+/* default filename is /proc/self/mountinfo
  */
-int mnt_table_parse_mtab(struct libmnt_table *tb, const char *filename)
+int __mnt_table_parse_mtab(struct libmnt_table *tb, const char *filename,
+			   struct libmnt_table *u_tb)
 {
-	int rc;
-	const char *utab = NULL;
-	struct libmnt_table *u_tb;
+	int rc = 0, priv_utab = 0;
+	int explicit_file = filename ? 1 : 0;
 
 	assert(tb);
 
+	if (filename)
+		DBG(TAB, ul_debugobj(tb, "%s requested as mtab", filename));
+
+#ifdef USE_LIBMOUNT_SUPPORT_MTAB
 	if (mnt_has_regular_mtab(&filename, NULL)) {
 
-		DBG(TAB, mnt_debug_h(tb, "force %s usage", filename));
+		DBG(TAB, ul_debugobj(tb, "force mtab usage [filename=%s]", filename));
 
 		rc = mnt_table_parse_file(tb, filename);
+
+		/*
+		 * If @filename forces us to read from /proc then also read
+		 * utab file to merge userspace mount options.
+		 */
+		if (rc == 0 && is_mountinfo(tb))
+			goto read_utab;
+
 		if (!rc)
 			return 0;
 		filename = NULL;	/* failed */
-	}
+	} else
+		filename = NULL;	/* mtab useless */
+#endif
 
-	/*
-	 * useless /etc/mtab
-	 * -- read kernel information from /proc/self/mountinfo
-	 */
-	tb->fmt = MNT_FMT_MOUNTINFO;
-	rc = mnt_table_parse_file(tb, _PATH_PROC_MOUNTINFO);
+	if (!filename || strcmp(filename, _PATH_PROC_MOUNTINFO) == 0) {
+		filename = _PATH_PROC_MOUNTINFO;
+		tb->fmt = MNT_FMT_MOUNTINFO;
+		DBG(TAB, ul_debugobj(tb, "mtab parse: #1 read mountinfo"));
+	} else
+		tb->fmt = MNT_FMT_GUESS;
+
+	rc = mnt_table_parse_file(tb, filename);
 	if (rc) {
+		if (explicit_file)
+			return rc;
+
 		/* hmm, old kernel? ...try /proc/mounts */
 		tb->fmt = MNT_FMT_MTAB;
 		return mnt_table_parse_file(tb, _PATH_PROC_MOUNTS);
 	}
 
+	if (!is_mountinfo(tb))
+		return 0;
+#ifdef USE_LIBMOUNT_SUPPORT_MTAB
+read_utab:
+#endif
+	DBG(TAB, ul_debugobj(tb, "mtab parse: #2 read utab"));
+
 	if (mnt_table_get_nents(tb) == 0)
 		return 0;			/* empty, ignore utab */
 	/*
 	 * try to read the user specific information from /run/mount/utabs
 	 */
-	utab = mnt_get_utab_path();
-	if (!utab || is_file_empty(utab))
-		return 0;
+	if (!u_tb) {
+		const char *utab = mnt_get_utab_path();
 
-	u_tb = mnt_new_table();
-	if (!u_tb)
-		return -ENOMEM;
+		if (!utab || is_file_empty(utab))
+			return 0;
 
-	u_tb->fmt = MNT_FMT_UTAB;
-	mnt_table_set_parser_fltrcb(u_tb, tb->fltrcb, tb->fltrcb_data);
+		u_tb = mnt_new_table();
+		if (!u_tb)
+			return -ENOMEM;
 
-	if (mnt_table_parse_file(u_tb, utab) == 0) {
+		u_tb->fmt = MNT_FMT_UTAB;
+		mnt_table_set_parser_fltrcb(u_tb, tb->fltrcb, tb->fltrcb_data);
+
+		rc = mnt_table_parse_file(u_tb, utab);
+		priv_utab = 1;
+	}
+
+	DBG(TAB, ul_debugobj(tb, "mtab parse: #3 merge utab"));
+
+	if (rc == 0) {
 		struct libmnt_fs *u_fs;
 		struct libmnt_iter itr;
 
@@ -1114,6 +1314,30 @@ int mnt_table_parse_mtab(struct libmnt_table *tb, const char *filename)
 			mnt_table_merge_user_fs(tb, u_fs);
 	}
 
-	mnt_unref_table(u_tb);
+
+	if (priv_utab)
+		mnt_unref_table(u_tb);
 	return 0;
 }
+/**
+ * mnt_table_parse_mtab:
+ * @tb: table
+ * @filename: overwrites default or NULL
+ *
+ * The default filename is /proc/self/mountinfo. If the mount table is a
+ * mountinfo file then /run/mount/utabs is parsed too and both files are merged
+ * to the one libmnt_table.
+ *
+ * If libmount is compiled with classic mtab file support, and the /etc/mtab is
+ * a regular file then this file is parsed.
+ *
+ * It's strongly recommended to use NULL as a @filename to keep code portable.
+ *
+ * See also mnt_table_set_parser_errcb().
+ *
+ * Returns: 0 on success or negative number in case of error.
+ */
+int mnt_table_parse_mtab(struct libmnt_table *tb, const char *filename)
+{
+	return __mnt_table_parse_mtab(tb, filename, NULL);
+}
